# used for delayed evaluation of typing until python 3.11 becomes mainstream
from __future__ import annotations

import os
import threading
import time
import signal
from abc import ABC, abstractmethod
from copy import copy

import numpy as np
from trimesh import Trimesh
from typing_extensions import List, Optional, Dict, Tuple, Callable, TYPE_CHECKING, Union, Type, deprecated

import pycrap
from pycrap.ontologies import PhysicalObject, Robot, Floor, Apartment
from pycrap.ontologies.crax.rules import HierarchicalContainment, CRAXRule
from pycrap.ontology_wrapper import OntologyWrapper
from ..cache_manager import CacheManager
from ..config.world_conf import WorldConfig
from ..datastructures.dataclasses import (Color, AxisAlignedBoundingBox, CollisionCallbacks,
                                          MultiBody, VisualShape, BoxVisualShape, CylinderVisualShape,
                                          SphereVisualShape,
                                          CapsuleVisualShape, PlaneVisualShape, MeshVisualShape,
                                          ObjectState, WorldState, ClosestPointsList,
                                          ContactPointsList, VirtualMobileBaseJoints, RotatedBoundingBox, RayResult)
from ..datastructures.enums import JointType, WorldMode, Arms, AdjacentBodyMethod as ABM
from ..datastructures.pose import PoseStamped, TransformStamped, Point
from ..datastructures.world_entity import PhysicalBody, WorldEntity
from ..failures import ProspectionObjectNotFound, ObjectNotFound
from ..local_transformer import LocalTransformer
from ..robot_description import RobotDescription
from ..ros import Time
from ..ros import logwarn
from ..validation.goal_validator import (GoalValidator,
                                         validate_joint_position, validate_multiple_joint_positions,
                                         validate_object_pose, validate_multiple_object_poses)
from ..world_concepts.constraints import Constraint
from ..world_concepts.event import Event

if TYPE_CHECKING:
    from ..world_concepts.world_object import Object
    from ..description import Link, Joint, ObjectDescription
    from ..object_descriptors.generic import ObjectDescription as GenericObjectDescription


class World(WorldEntity, ABC):
    """
    The World Class represents the physics Simulation and belief state, it is the main interface for reasoning about
    the World. This is implemented as a singleton, the current World can be accessed via the static variable
    current_world which is managed by the World class itself.
    """

    conf: Type[WorldConfig] = WorldConfig
    """
    The configurations of the world, the default configurations are defined in world_conf.py in the config folder.
    """

    current_world: Optional[World] = None
    """
    Global reference to the currently used World, usually this is the
    graphical one. However, if you are inside a UseProspectionWorld() environment the current_world points to the
    prospection world. In this way you can comfortably use the current_world, which should point towards the World
    used at the moment.
    """

    robot: Optional[Object] = None
    """
    Global reference to the spawned Object that represents the robot. The robot is identified by checking the name in
     the URDF with the name of the URDF on the parameter server. 
    """

    environment: Optional[Object] = None
    """
    Global reference to the spawned Object that represents the environment. The environment is identified by checking the name in
     the URDF with the name of the URDF on the parameter server. 
    """

    cache_manager: CacheManager = CacheManager(conf.cache_dir, [conf.resources_path], False)
    """
    Global reference for the cache manager, this is used to cache the description files of the robot and the objects.
    """

    ontology: Optional[OntologyWrapper] = None
    """
    The ontology of this world.
    """

    def __init__(self, mode: WorldMode = WorldMode.DIRECT, is_prospection: bool = False, clear_cache: bool = False,
                 id_: int = -1):
        """
        Create a new simulation, the mode decides if the simulation should be a rendered window or just run in the
        background. There can only be one rendered simulation.
        The World object also initializes the Events for attachment, detachment and for manipulating the world.

        :param mode: Can either be "GUI" for rendered window or "DIRECT" for non-rendered. The default parameter is
         "GUI"
        :param is_prospection: For internal usage, decides if this World should be used as a prospection world.
        :param clear_cache: Whether to clear the cache directory.
        :param id_: The unique id of the world.
        """
        self.is_prospection_world: bool = is_prospection
        if not is_prospection:
            self.ontology = OntologyWrapper()
        else:
            self.ontology = None
        WorldEntity.__init__(self, id_, self, concept=pycrap.ontologies.World)

        self.latest_state_id: Optional[int] = None

        if clear_cache or (self.conf.clear_cache_at_start and not self.cache_manager.cache_cleared):
            self.cache_manager.clear_cache()

        GoalValidator.raise_error = self.conf.raise_goal_validator_error

        if World.current_world is None:
            World.current_world = self

        self.object_lock: threading.Lock = threading.Lock()

        self._init_world(mode)

        self.objects: List[Object] = []
        # List of all Objects in the World

        self._init_and_sync_prospection_world()

        self.local_transformer = LocalTransformer()
        self._update_local_transformer_worlds()

        self.mode: WorldMode = mode

        self.coll_callbacks: Dict[Tuple[Object, Object], CollisionCallbacks] = {}

        self._init_events()

        self._current_state: Optional[WorldState] = None

        self.original_state_id = self.save_state()

        self.on_add_object_callbacks: List[Callable[[Object], None]] = []

        self._set_world_rules()

        signal.signal(signal.SIGINT, self.signal_handler)

    def _set_world_rules(self):
        """
        Create the rules for the world.
        """
        if self.is_prospection_world:
            return
        HierarchicalContainment()

    @property
    def rules(self) -> List[CRAXRule]:
        """
        Return the rules of the world.
        """
        return list(CRAXRule.all_rules[self.ontology].values())

    @staticmethod
    def update_containment_for(bodies: List[PhysicalBody],
                               candidate_selection_method: ABM = ABM.ClosestPoints) \
            -> List[PhysicalBody]:
        """
        Update the containment for the given bodies by checking if they are contained in other bodies.

        :param bodies: The bodies to update the containment for.
        :param candidate_selection_method: The method to select the candidate bodies for containment update.
        :return: The updated bodies.
        """
        checked_bodies: List[PhysicalBody] = []
        for body in bodies:
            body.update_containment(excluded_bodies=checked_bodies,
                                    candidate_selection_method=candidate_selection_method)
            checked_bodies.append(body)
        return checked_bodies

    @property
    def parent_entity(self) -> Optional[WorldEntity]:
        """
        Return the parent entity of this entity, in this case it is None as the World is the top level entity.
        """
        return None

    @property
    def name(self) -> str:
        """
        Return the name of the world, which is the name of the implementation class (e.g. BulletWorld).
        """
        return self.__class__.__name__

    def get_body_convex_hull(self, body: PhysicalBody) -> Trimesh:
        """
        :param body: The body object.
        :return: The convex hull of the body as a Geometry3D object.
        """
        raise NotImplementedError

    def add_callback_on_add_object(self, callback: Callable[[Object], None]) -> None:
        """
        Add a callback that is called when an object is added to the world.

        :param callback: The callback.
        """
        self.on_add_object_callbacks.append(callback)

    def remove_callback_on_add_object(self, callback: Callable[[Object], None]) -> None:
        """
        Remove a callback that is called when an object is added to the world.

        :param callback: The callback.
        """
        self.on_add_object_callbacks.remove(callback)

    @classmethod
    def get_cache_dir(cls) -> str:
        """
        Return the cache directory.
        """
        return cls.cache_manager.cache_dir

    def add_object(self, obj: Object) -> None:
        """
        Add an object to the world.

        :param obj: The object to be added.
        """
        self.object_lock.acquire()
        self.objects.append(obj)
        self.add_object_to_original_state(obj)
        self.object_lock.release()
        self.invoke_on_add_object_callbacks(obj)

    def invoke_on_add_object_callbacks(self, obj: Object) -> None:
        """
        Invoke the object added callbacks.

        :param obj: The object.
        """
        for callback in self.on_add_object_callbacks:
            callback(obj)

    @property
    def robot_description(self) -> RobotDescription:
        """
        Return the current robot description.
        """
        return RobotDescription.current_robot_description

    @property
    def robot_has_actuators(self) -> bool:
        """
        Return whether the robot has actuators.
        """
        return self.robot_description.has_actuators

    def get_actuator_for_joint(self, joint: Joint) -> str:
        """
        Get the actuator name for a given joint.
        """
        return self.robot_joint_actuators[joint.name]

    def joint_has_actuator(self, joint: Joint) -> bool:
        """
        Return whether the joint has an actuator.
        """
        return joint.name in self.robot_joint_actuators

    @property
    def robot_joint_actuators(self) -> Dict[str, str]:
        """
        Return the joint actuators of the robot.
        """
        return self.robot_description.joint_actuators

    def check_object_exists(self, obj: Object) -> bool:
        """
        Check if the object exists in the simulator.

        :param obj: The object to check.
        :return: True if the object is in the world, False otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def _init_world(self, mode: WorldMode):
        """
        Initialize the physics simulation.
        """
        raise NotImplementedError

    def _init_events(self):
        """
        Initialize dynamic events that can be used to react to changes in the World.
        """
        self.detachment_event: Event = Event()
        self.attachment_event: Event = Event()
        self.manipulation_event: Event = Event()

    def _init_and_sync_prospection_world(self):
        """
        Initialize the prospection world and the synchronization between the main and the prospection world.
        """
        self._init_prospection_world()
        self._sync_prospection_world()

    def _update_local_transformer_worlds(self):
        """
        Update the local transformer worlds with the current world and prospection world.
        """
        self.local_transformer.world = self
        self.local_transformer.prospection_world = self.prospection_world

    def _init_prospection_world(self):
        """
        Initialize the prospection world, if this is a prospection world itself it will not create another prospection,
        world, but instead set the prospection world to None, else it will create a prospection world.
        """
        if self.is_prospection_world:  # then no need to add another prospection world
            self.prospection_world = None
        else:
            self.prospection_world: World = self.__class__(is_prospection=True)

    def _sync_prospection_world(self):
        """
        Synchronize the prospection world with the main world, this means that every object in the main world will be
        added to the prospection world and vice versa.
        """
        if self.is_prospection_world:  # then no need to add another prospection world
            self.world_sync = None
        else:
            self.world_sync: WorldSync = WorldSync(self, self.prospection_world)
            self.world_sync.start()

    def preprocess_object_file_and_get_its_cache_path(self, path: str, ignore_cached_files: bool,
                                                      description: ObjectDescription, name: str,
                                                      scale_mesh: Optional[float] = None,
                                                      mesh_transform: Optional[TransformStamped] = None,
                                                      color: Optional[Color] = None) -> str:
        """
        Update the cache directory with the given object.

        :param path: The path to the object.
        :param ignore_cached_files: If the cached files should be ignored.
        :param description: The object description.
        :param name: The name of the object.
        :param scale_mesh: The scale of the mesh.
        :param mesh_transform: The mesh transform to apply to the mesh.
        :param color: The color of the object.
        :return: The path of the cached object.
        """
        return self.cache_manager.update_cache_dir_with_object(path, ignore_cached_files, description, name,
                                                               scale_mesh, mesh_transform, color)

    @property
    def simulation_time_step(self):
        """
        The time step of the simulation in seconds.
        """
        return 1 / self.__class__.conf.simulation_frequency

    @abstractmethod
    def load_object_and_get_id(self, path: Optional[str] = None, pose: Optional[PoseStamped] = None,
                               obj_type: Optional[Type[PhysicalObject]] = None) -> int:
        """
        Load a description file (e.g. URDF) at the given pose and returns the id of the loaded object.

        :param path: The path to the description file, if None the description file is assumed to be already loaded.
        :param pose: The pose at which the object should be loaded.
        :param obj_type: The type of the object.
        :return: The id of the loaded object.
        """
        pass

    def load_generic_object_and_get_id(self, description: GenericObjectDescription,
                                       pose: Optional[PoseStamped] = None) -> int:
        """
        Create a visual and collision box in the simulation and returns the id of the loaded object.

        :param description: The object description.
        :param pose: The pose at which the object should be loaded.
        """
        raise NotImplementedError

    def get_object_names(self) -> List[str]:
        """
        Return the names of all objects in the World.

        :return: A list of object names.
        """
        return [obj.name for obj in self.objects]

    def get_object_by_name(self, name: str) -> Optional[Object]:
        """
        Return the object with the given name. If there is no object with the given name, None is returned.

        :param name: The name of the returned Objects.
        :return: The object with the given name, if there is one.
        """

        matching_objects = list(filter(lambda obj: obj.name == name, self.objects))
        return matching_objects[0] if len(matching_objects) > 0 else None

    def get_object_by_type(self, obj_type: PhysicalObject) -> List[Object]:
        """
        Return a list of all Objects which have the type 'obj_type'.

        :param obj_type: The type of the returned Objects.
        :return: A list of all Objects that have the type 'obj_type'.
        """
        return list(filter(lambda obj: obj.obj_type == obj_type, self.objects))

    def get_object_by_id(self, obj_id: int) -> Object:
        """
        Return the single Object that has the unique id.

        :param obj_id: The unique id for which the Object should be returned.
        :return: The Object with the id 'id'.
        """
        return list(filter(lambda obj: obj.id == obj_id, self.objects))[0]

    def get_scene_objects(self) -> List[Object]:
        """
        :return: A list of all objects in the world except the robot, floor, and apartment.
        """
        return [obj for obj in self.objects if obj.obj_type not in {Robot, Floor, Apartment}]

    def remove_visual_object(self, obj_id: int) -> bool:
        """
        Remove the object with the given id from the world, and saves a new original state for the world.

        :param obj_id: The unique id of the object to be removed.
        :return: Whether the object was removed successfully.
        """

        removed = self._remove_visual_object(obj_id)
        if removed:
            self.update_simulator_state_id_in_original_state()
        else:
            logwarn(f"Object with id {obj_id} could not be removed.")
        return removed

    @abstractmethod
    def _remove_visual_object(self, obj_id: int) -> bool:
        """
        Remove the visual object with the given id from the world, and update the simulator state in the original state.

        :param obj_id: The unique id of the visual object to be removed.
        :return: Whether the object was removed successfully.
        """
        pass

    @abstractmethod
    def remove_object_from_simulator(self, obj: Object) -> bool:
        """
        Remove an object from the physics simulator.

        :param obj: The object to be removed.
        :return: Whether the object was removed successfully.
        """
        pass

    def remove_object(self, obj: Object) -> None:
        """
        Remove this object from the current world.
        For the object to be removed it has to be detached from all objects it
        is currently attached to. After this is done a call to world remove object is done
        to remove this Object from the simulation/world.

        :param obj: The object to be removed.
        """
        self.object_lock.acquire()

        obj.detach_all()

        if self.remove_object_from_simulator(obj):
            self.objects.remove(obj)
            self.remove_object_from_original_state(obj)

            if World.robot == obj and not self.is_prospection_world:
                World.robot = None
        else:
            logwarn(f"Object {obj.name} could not be removed from the simulator, but all attachments were removed")

        self.object_lock.release()

    def remove_object_from_original_state(self, obj: Object) -> None:
        """
        Remove an object from the original state of the world.

        :param obj: The object to be removed.
        """
        self.original_state.object_states.pop(obj.name)
        self.update_simulator_state_id_in_original_state(use_same_id=True)

    def add_object_to_original_state(self, obj: Object) -> None:
        """
        Add an object to the original state of the world.

        :param obj: The object to be added.
        """
        self.original_state.object_states[obj.name] = obj.current_state
        self.update_simulator_state_id_in_original_state()

    def add_fixed_constraint(self, parent_link: Link, child_link: Link,
                             child_to_parent_transform: TransformStamped) -> int:
        """
        Create a fixed joint constraint between the given parent and child links,
        the joint frame will be at the origin of the child link frame, and would have the same orientation
        as the child link frame.

        :param parent_link: The constrained link of the parent object.
        :param child_link: The constrained link of the child object.
        :param child_to_parent_transform: The transform from the child link frame to the parent link frame.
        :return: The unique id of the created constraint.
        """

        constraint = Constraint(parent_link=parent_link,
                                child_link=child_link,
                                _type=JointType.FIXED,
                                axis_in_child_frame=Point(x=0, y=0, z=0),
                                constraint_to_parent=child_to_parent_transform,
                                child_to_constraint=TransformStamped.from_list(frame=child_link.tf_frame)
                                )
        constraint_id = self.add_constraint(constraint)
        return constraint_id

    @abstractmethod
    def add_constraint(self, constraint: Constraint) -> int:
        """
        Add a constraint between two objects links so that they become attached for example.

        :param constraint: The constraint data used to create the constraint.
        """
        pass

    @abstractmethod
    def remove_constraint(self, constraint_id) -> None:
        """
        Remove a constraint by its ID.

        :param constraint_id: The unique id of the constraint to be removed.
        """
        pass

    def get_joint_position(self, joint: Joint) -> float:
        """
        Wrapper for :meth:`_get_joint_position` that return 0.0 for a joint if it is in the ignore joints list.
        """
        if joint.object.is_a_robot and joint.name in self.robot_description.ignore_joints:
            return 0.0
        return self._get_joint_position(joint)

    @abstractmethod
    def _get_joint_position(self, joint: Joint) -> float:
        """
        Get the position of a joint of an articulated object

        :param joint: The joint to get the position for.
        :return: The joint position as a float.
        """
        pass

    @abstractmethod
    def get_object_joint_names(self, obj: Object) -> List[str]:
        """
        Return the names of all joints of this object.

        :param obj: The object.
        :return: A list of joint names.
        """
        pass

    @abstractmethod
    def get_link_pose(self, link: Link) -> PoseStamped:
        """
        Get the pose of a link of an articulated object with respect to the world frame.

        :param link: The link as a AbstractLink object.
        :return: The pose of the link as a Pose object.
        """
        pass

    @abstractmethod
    def get_multiple_link_poses(self, links: List[Link]) -> Dict[str, PoseStamped]:
        """
        Get the poses of multiple links of an articulated object with respect to the world frame.

        :param links: The links as a list of AbstractLink objects.
        :return: A dictionary with link names as keys and Pose objects as values.
        """
        pass

    @abstractmethod
    def get_link_position(self, link: Link) -> List[float]:
        """
        Get the position of a link of an articulated object with respect to the world frame.

        :param link: The link as a AbstractLink object.
        :return: The position of the link as a list of floats.
        """
        pass

    @abstractmethod
    def get_link_orientation(self, link: Link) -> List[float]:
        """
        Get the orientation of a link of an articulated object with respect to the world frame.

        :param link: The link as a AbstractLink object.
        :return: The orientation of the link as a list of floats.
        """
        pass

    @abstractmethod
    def get_multiple_link_positions(self, links: List[Link]) -> Dict[str, List[float]]:
        """
        Get the positions of multiple links of an articulated object with respect to the world frame.

        :param links: The links as a list of AbstractLink objects.
        :return: A dictionary with link names as keys and lists of floats as values.
        """
        pass

    @abstractmethod
    def get_multiple_link_orientations(self, links: List[Link]) -> Dict[str, List[float]]:
        """
        Get the orientations of multiple links of an articulated object with respect to the world frame.

        :param links: The links as a list of AbstractLink objects.
        :return: A dictionary with link names as keys and lists of floats as values.
        """
        pass

    @abstractmethod
    def get_object_link_names(self, obj: Object) -> List[str]:
        """
        Return the names of all links of this object.

        :param obj: The object.
        :return: A list of link names.
        """
        pass

    def simulate(self, seconds: float, real_time: Optional[bool] = False,
                 func: Optional[Callable[[], None]] = None) -> None:
        """
        Simulate Physics in the World for a given amount of seconds. Usually this simulation is faster than real
        time. By setting the 'real_time' parameter this simulation is slowed down such that the simulated time is equal
        to real time.

        :param seconds: The amount of seconds that should be simulated.
        :param real_time: If the simulation should happen in real time or faster.
        :param func: A function that should be called during the simulation
        """
        self.set_realtime(real_time)
        for i in range(0, int(seconds * self.conf.simulation_frequency)):
            curr_time = Time().now()
            self.step(func)
            for objects, callbacks in self.coll_callbacks.items():
                contact_points = self.get_contact_points_between_two_bodies(objects[0], objects[1])
                if len(contact_points) > 0:
                    callbacks.on_collision_cb()
                elif callbacks.no_collision_cb is not None:
                    callbacks.no_collision_cb()
            if real_time:
                loop_time = Time().now() - curr_time
                time_diff = self.simulation_time_step - loop_time.to_sec()
                time.sleep(max(0, time_diff))

    @abstractmethod
    def get_object_pose(self, obj: Object) -> PoseStamped:
        """
        Get the pose of an object in the world frame from the current object pose in the simulator.

        :param obj: The object.
        """
        pass

    @abstractmethod
    def get_multiple_object_poses(self, objects: List[Object]) -> Dict[str, PoseStamped]:
        """
        Get the poses of multiple objects in the world frame from the current object poses in the simulator.

        :param objects: The objects.
        """
        pass

    @abstractmethod
    def get_multiple_object_positions(self, objects: List[Object]) -> Dict[str, List[float]]:
        """
        Get the positions of multiple objects in the world frame from the current object poses in the simulator.

        :param objects: The objects.
        """
        pass

    @abstractmethod
    def get_object_position(self, obj: Object) -> List[float]:
        """
        Get the position of an object in the world frame from the current object pose in the simulator.

        :param obj: The object.
        """
        pass

    @abstractmethod
    def get_multiple_object_orientations(self, objects: List[Object]) -> Dict[str, List[float]]:
        """
        Get the orientations of multiple objects in the world frame from the current object poses in the simulator.

        :param objects: The objects.
        """
        pass

    @abstractmethod
    def get_object_orientation(self, obj: Object) -> List[float]:
        """
        Get the orientation of an object in the world frame from the current object pose in the simulator.

        :param obj: The object.
        """
        pass

    @property
    def robot_virtual_joints(self) -> List[Joint]:
        """
        The virtual joints of the robot.
        """
        return [self.robot.joints[name] for name in self.robot_virtual_joints_names]

    @property
    def robot_virtual_joints_names(self) -> List[str]:
        """
        The names of the virtual joints of the robot.
        """
        return self.robot_description.virtual_mobile_base_joints.names

    def get_robot_mobile_base_joints(self) -> VirtualMobileBaseJoints:
        """
        Get the mobile base joints of the robot.

        :return: The mobile base joints.
        """
        return self.robot_description.virtual_mobile_base_joints

    @abstractmethod
    def perform_collision_detection(self) -> None:
        """
        Check for collisions between all objects in the World and updates the contact points.
        """
        pass

    @deprecated("Use get_body_contact_points instead")
    def get_object_contact_points(self, obj: Object) -> ContactPointsList:
        """
        Same as :meth:`get_body_contact_points` but with objects instead of any type of bodies.
        """
        return self.get_body_contact_points(obj)

    @abstractmethod
    def get_body_contact_points(self, body: PhysicalBody) -> ContactPointsList:
        """
        Return the contact points of a body with all other bodies in the world.

        :param body: The body.
        """
        pass

    @deprecated("Use get_contact_points_between_two_bodies instead")
    def get_contact_points_between_two_objects(self, obj1: Object, obj2: Object) -> ContactPointsList:
        """
        Same as :meth:`get_contact_points_between_two_bodies` but with objects instead of any type of bodies.
        """
        return self.get_contact_points_between_two_bodies(obj1, obj2)

    @abstractmethod
    def get_contact_points_between_two_bodies(self, body_1: PhysicalBody, body_2: PhysicalBody) -> ContactPointsList:
        """
        Return a list of contact points between two bodies.

        :param body_1: The first body.
        :param body_2: The second body.
        :return: A list of all contact points between the two bodies.
        """
        pass

    @deprecated("Use get_contact_points_between_two_bodies instead")
    def get_contact_points_between_two_objects(self, obj1: Object, obj2: Object) -> ContactPointsList:
        """
        Same as :meth:`get_contact_points_between_two_bodies` but with objects instead of any type of bodies.
        """
        return self.get_contact_points_between_two_bodies(obj1, obj2)

    def get_body_closest_points(self, body: PhysicalBody, max_distance: float) -> ClosestPointsList:
        """
        Return the closest points of this body with all other bodies in the world.

        :param body: The body.
        :param max_distance: The maximum allowed distance between the points.
        :return: A list of the closest points.
        """
        all_obj_closest_points = [self.get_closest_points_between_two_bodies(body, other_body, max_distance)
                                  for other_body in self.objects if other_body != body]
        return ClosestPointsList([point for closest_points in all_obj_closest_points for point in closest_points])

    def get_closest_points_between_two_bodies(self, body_a: PhysicalBody, body_b: PhysicalBody, max_distance: float) \
            -> ClosestPointsList:
        """
        Return the closest points between two objects.

        :param body_a: The first body.
        :param body_b: The second body.
        :param max_distance: The maximum distance between the points.
        :return: A list of the closest points.
        """
        raise NotImplementedError

    def reset_joint_position(self, joint: Joint, joint_position: float) -> bool:
        """
        Wrapper around :meth:`_reset_joint_position` that checks if the joint should be ignored.
        """
        if joint.object.is_a_robot and self.robot_description.ignore_joints:
            if joint.name in self.robot_description.ignore_joints:
                return True
        return self._reset_joint_position(joint, joint_position)

    @validate_joint_position
    @abstractmethod
    def _reset_joint_position(self, joint: Joint, joint_position: float) -> bool:
        """
        Reset the joint position instantly without physics simulation

        .. note::
           It is recommended to use the validate_joint_position decorator to validate the joint position for
           the implementation of this method.

        :param joint: The joint to reset the position for.
        :param joint_position: The new joint pose.
        :return: True if the reset was successful, False otherwise
        """
        pass

    def set_multiple_joint_positions(self, joint_positions: Dict[Joint, float]) -> bool:
        """
        Wrapper around :meth:`_set_multiple_joint_positions` that checks if any of the joints should be ignored.
        """
        filtered_joint_positions = copy(joint_positions)
        for joint, position in joint_positions.items():
            if joint.name in self.robot_description.ignore_joints:
                filtered_joint_positions.pop(joint)
        return self._set_multiple_joint_positions(filtered_joint_positions)

    @validate_multiple_joint_positions
    @abstractmethod
    def _set_multiple_joint_positions(self, joint_positions: Dict[Joint, float]) -> bool:
        """
        Set the positions of multiple joints of an articulated object.

        .. note::
           It is recommended to use the validate_multiple_joint_positions decorator to validate the
           joint positions for the implementation of this method.

        :param joint_positions: A dictionary with joint objects as keys and joint positions as values.
        :return: True if the set was successful, False otherwise.
        """
        pass

    def get_multiple_joint_positions(self, joints: List[Joint]) -> Dict[str, float]:
        """
        Wrapper around :meth:`_get_multiple_joint_positions` that checks if any of the joints should be ignored.
        """
        filtered_joints = [joint for joint in joints if not joint.object.is_a_robot or
                           joint.name not in self.robot_description.ignore_joints]
        joint_positions = self._get_multiple_joint_positions(filtered_joints)
        joint_positions.update({joint.name: 0.0 for joint in joints if joint not in filtered_joints})
        return joint_positions

    @abstractmethod
    def _get_multiple_joint_positions(self, joints: List[Joint]) -> Dict[str, float]:
        """
        Get the positions of multiple joints of an articulated object.

        :param joints: The joints as a list of Joint objects.
        """
        pass

    @validate_object_pose
    @abstractmethod
    def reset_object_base_pose(self, obj: Object, pose: PoseStamped) -> bool:
        """
        Reset the world position and orientation of the base of the object instantaneously,
        not through physics simulation. (x,y,z) position vector and (x,y,z,w) quaternion orientation.

        .. note::
           It is recommended to use the validate_object_pose decorator to validate the object pose for the
           implementation of this method.

        :param obj: The object.
        :param pose: The new pose as a Pose object.
        :return: True if the reset was successful, False otherwise.
        """
        pass

    @validate_multiple_object_poses
    @abstractmethod
    def reset_multiple_objects_base_poses(self, objects: Dict[Object, PoseStamped]) -> bool:
        """
        Reset the world position and orientation of the base of multiple objects instantaneously,
        not through physics simulation. (x,y,z) position vector and (x,y,z,w) quaternion orientation.

        :param objects: A dictionary with objects as keys and poses as values.
        :return: True if the reset was successful, False otherwise.
        """
        pass

    @abstractmethod
    def step(self, func: Optional[Callable[[], None]] = None, step_seconds: Optional[float] = None) -> None:
        """
        Step the world simulation using forward dynamics.

        :param func: An optional function to be called during the step.
        :param step_seconds: The amount of seconds to step the simulation if None the simulation is stepped by the
        simulation time step.
        """
        pass

    def get_arm_tool_frame_link(self, arm: Arms) -> Link:
        """
        Get the tool frame link of the arm of the robot.

        :param arm: The arm for which the tool frame link should be returned.
        :return: The tool frame link of the arm.
        """
        ee_link_name = self.robot_description.get_arm_tool_frame(arm)
        return self.robot.get_link(ee_link_name)

    @abstractmethod
    def set_link_color(self, link: Link, rgba_color: Color):
        """
        Change the rgba_color of a link of this object, the rgba_color has to be given as Color object.

        :param link: The link which should be colored.
        :param rgba_color: The rgba_color as Color object with RGBA values between 0 and 1.
        """
        pass

    @abstractmethod
    def get_link_color(self, link: Link) -> Color:
        """
        :param link: The link for which the rgba_color should be returned.
        :return: The rgba_color as Color object with RGBA values between 0 and 1.
        """
        pass

    @abstractmethod
    def get_colors_of_object_links(self, obj: Object) -> Dict[str, Color]:
        """
        :param obj: The object
        :return: The RGBA colors of each link in the object as a dictionary from link name to rgba_color.
        """
        pass

    def get_object_axis_aligned_bounding_box(self, obj: Object) -> AxisAlignedBoundingBox:
        """
        :param obj: The object for which the bounding box should be returned.
        :return: the axis aligned bounding box of this object. The return of this method are two points in
        world coordinate frame which define a bounding box.
        """
        raise NotImplementedError()

    def get_object_rotated_bounding_box(self, obj: Object) -> RotatedBoundingBox:
        """
        :param obj: The object for which the bounding box should be returned.
        :return: the rotated bounding box of this object. The return of this method are two points in
        world coordinate frame which define a bounding box.
        """
        raise NotImplementedError()

    def get_link_axis_aligned_bounding_box(self, link: Link) -> AxisAlignedBoundingBox:
        """
        :param link: The link for which the bounding box should be returned.
        :return: The axis aligned bounding box of the link. The return of this method are two points in
        world coordinate frame which define a bounding box.
        """
        raise NotImplementedError()

    def get_link_rotated_bounding_box(self, link: Link) -> RotatedBoundingBox:
        """
        :param link: The link for which the bounding box should be returned.
        :return: The rotated bounding box of the link. The return of this method are two points in
        world coordinate frame which define a bounding box.
        """
        raise NotImplementedError()

    @abstractmethod
    def set_realtime(self, real_time: bool) -> None:
        """
        Enable the real time simulation of Physics in the World. By default, this is disabled and Physics is only
        simulated to reason about it.

        :param real_time: Whether the World should simulate Physics in real time.
        """
        pass

    @abstractmethod
    def set_gravity(self, gravity_vector: List[float]) -> None:
        """
        Set the gravity that is used in the World. By default, it is set to the gravity on earth ([0, 0, -9.8]).
         Gravity is given as a vector in x,y,z. Gravity is only applied while simulating Physic.

        :param gravity_vector: The gravity vector that should be used in the World.
        """
        pass

    def set_robot_if_not_set(self, robot: Object) -> None:
        """
        Set the robot if it is not set yet.

        :param robot: The Object reference to the Object representing the robot.
        """
        if not self.robot_is_set():
            self.set_robot(robot)

    @staticmethod
    def set_robot(robot: Union[Object, None]) -> None:
        """
        Set the global variable for the robot Object This should be set on spawning the robot.

        :param robot: The Object reference to the Object representing the robot.
        """
        World.robot = robot

    @staticmethod
    def robot_is_set() -> bool:
        """
        Return whether the robot has been set or not.

        :return: True if the robot has been set, False otherwise.
        """
        return World.robot is not None

    def exit(self, remove_saved_states: bool = True) -> None:
        """
        Close the World as well as the prospection world, also collects any other thread that is running.

        :param remove_saved_states: Whether to remove the saved states.
        """
        self.reset_world(remove_saved_states)
        self.remove_all_objects()
        self.exit_prospection_world_if_exists()
        self.disconnect_from_physics_server()
        self.reset_robot()
        self.join_threads()
        if self.ontology:
            self.ontology.destroy_individuals()
        if World.current_world == self:
            World.current_world = None

    def exit_prospection_world_if_exists(self) -> None:
        """
        Exit the prospection world if it exists.
        """
        if self.prospection_world:
            self.terminate_world_sync()
            self.prospection_world.exit()

    def signal_handler(self, sig, frame) -> None:
        """
        Signal handler for graceful exit of the world when a signal is received (e.g., SIGINT).
        """
        self.exit()
        print("Exiting World ...")
        exit(0)

    @abstractmethod
    def disconnect_from_physics_server(self) -> None:
        """
        Disconnect the world from the physics server.
        """
        pass

    def reset_current_world(self) -> None:
        """
        Reset the pose of every object in the World to the pose it was spawned in and sets every joint to 0.
        """
        for obj in self.objects:
            obj.set_pose(obj.original_pose)
            obj.set_multiple_joint_positions(dict(zip(list(obj.joint_names), [0] * len(obj.joint_names))))

    def reset_robot(self) -> None:
        """
        Set the robot class variable to None.
        """
        self.set_robot(None)

    @abstractmethod
    def join_threads(self) -> None:
        """
        Join any running threads. Useful for example when exiting the world.
        """
        pass

    def terminate_world_sync(self) -> None:
        """
        Terminate the world sync thread.
        """
        self.world_sync.terminate = True
        self.world_sync.join()

    def save_state(self, state_id: Optional[int] = None, use_same_id: bool = False, to_file: bool = False) -> int:
        """
        Return the id of the saved state of the World. The saved state contains the states of all the objects and
        the state of the physics simulator.

        :param state_id: The id of the saved state.
        :param use_same_id: Whether to use the same current state id for the new saved state.
        :param to_file: Whether to save the state to a file.
        :return: A unique id of the state
        """

        sim_state_id = self.save_physics_simulator_state(state_id=state_id, use_same_id=use_same_id)

        if state_id is None:
            if self.latest_state_id is None:
                self.latest_state_id = 0
            else:
                self.latest_state_id += 0 if use_same_id else 1
            state_id = self.latest_state_id

        self.save_objects_state(state_id)

        self._current_state = WorldState(self.object_states, sim_state_id)

        return super().save_state(state_id, self.get_cache_dir() if to_file else None)

    @property
    def current_state(self) -> WorldState:
        if self._current_state is None:
            simulator_state_id = None if not self.conf.use_physics_simulator_state else (
                self.save_physics_simulator_state(use_same_id=True))
            self._current_state = WorldState(self.object_states, simulator_state_id)
        return WorldState(self.object_states, self._current_state.simulator_state_id)

    @current_state.setter
    def current_state(self, state: WorldState) -> None:
        if self.current_state != state:
            if self.conf.use_physics_simulator_state:
                self.restore_physics_simulator_state(state.simulator_state_id)
                self.set_object_states_without_poses(state.object_states)
            else:
                for obj in self.objects:
                    self.get_object_by_name(obj.name).current_state = state.object_states[obj.name]

    def set_object_states_without_poses(self, states: Dict[str, ObjectState]) -> None:
        """
        Set the states of all objects in the World except the poses.

        :param states: A dictionary with the object id as key and the object state as value.
        """
        for obj_name, obj_state in states.items():
            obj = self.get_object_by_name(obj_name)
            obj.set_attachments(obj_state.attachments)
            obj.link_states = obj_state.link_states
            obj.joint_states = obj_state.joint_states
            if obj.name == self.robot.name and len(self.robot_virtual_joints) > 0:
                obj.set_mobile_robot_pose(obj_state.pose)

    @property
    def object_states(self) -> Dict[str, ObjectState]:
        """
        Return the states of all objects in the World.

        :return: A dictionary with the object id as key and the object state as value.
        """
        return {obj.name: obj.current_state for obj in self.objects}

    @object_states.setter
    def object_states(self, states: Dict[str, ObjectState]) -> None:
        """
        Set the states of all objects in the World.
        """
        for obj_name, obj_state in states.items():
            self.get_object_by_name(obj_name).current_state = obj_state

    def save_objects_state(self, state_id: int) -> None:
        """
        Save the state of all objects in the World according to the given state using the unique state id.

        :param state_id: The unique id representing the state.
        """
        for obj in self.objects:
            obj.save_state(state_id)

    @abstractmethod
    def save_physics_simulator_state(self, state_id: Optional[int] = None, use_same_id: bool = False) -> int:
        """
        Save the state of the physics simulator and returns the unique id of the state.

        :param state_id: The used specified unique id representing the state.
        :param use_same_id: If the same id should be used for the state.
        :return: The unique id representing the state.
        """
        pass

    @abstractmethod
    def remove_physics_simulator_state(self, state_id: int) -> None:
        """
        Remove the state of the physics simulator with the given id.

        :param state_id: The unique id representing the state.
        """
        pass

    @abstractmethod
    def restore_physics_simulator_state(self, state_id: int) -> None:
        """
        Restore the objects and environment state in the physics simulator according to
         the given state using the unique state id.

        :param state_id: The unique id representing the state.
        """
        pass

    def get_images_for_target(self,
                              target_pose: PoseStamped,
                              cam_pose: PoseStamped,
                              size: Optional[int] = 256) -> List[np.ndarray]:
        """
        Calculate the view and projection Matrix and returns 3 images:

        1. An RGB image
        2. A depth image
        3. A segmentation Mask, the segmentation mask indicates for every pixel the visible Object

        :param target_pose: The pose to which the camera should point.
        :param cam_pose: The pose of the camera.
        :param size: The height and width of the images in pixels.
        :return: A list containing an RGB and depth image as well as a segmentation mask, in this order.
        """
        pass

    def register_two_objects_collision_callbacks(self,
                                                 object_a: Object,
                                                 object_b: Object,
                                                 on_collision_callback: Callable,
                                                 on_collision_removal_callback: Optional[Callable] = None) -> None:
        """
        Register callback methods for contact between two Objects. There can be a callback for when the two Objects
        get in contact and, optionally, for when they are not in contact anymore.

        :param object_a: An object in the World
        :param object_b: Another object in the World
        :param on_collision_callback: A function that should be called if the objects are in contact
        :param on_collision_removal_callback: A function that should be called if the objects are not in contact
        """
        self.coll_callbacks[(object_a, object_b)] = CollisionCallbacks(on_collision_callback,
                                                                       on_collision_removal_callback)

    @classmethod
    def get_data_directories(cls) -> List[str]:
        """
        The resources directories where the objects, robots, and environments are stored.
        """
        return cls.cache_manager.data_directories

    @classmethod
    def add_resource_path(cls, path: str, prepend: bool = False) -> None:
        """
        Add a resource path in which the World will search for files. This resource directory is searched if an
        Object is spawned only with a filename.

        :param path: A path in the filesystem in which to search for files.
        :param prepend: Put the new path at the beginning of the list such that it is searched first.
        """
        if prepend:
            cls.cache_manager.data_directories = [path] + cls.cache_manager.data_directories
        else:
            cls.cache_manager.data_directories.append(path)

    @classmethod
    def remove_resource_path(cls, path: str) -> None:
        """
        Remove the given path from the data_directories list.

        :param path: The path to remove.
        """
        cls.cache_manager.data_directories.remove(path)

    @classmethod
    def change_cache_dir_path(cls, path: str) -> None:
        """
        Change the cache directory to the given path

        :param path: The new path for the cache directory.
        """
        cls.cache_manager.cache_dir = os.path.join(path, cls.conf.cache_dir_name)

    def get_prospection_object_for_object(self, obj: Object) -> Object:
        """
        Return the corresponding object from the prospection world for a given object in the main world.
         If the given Object is already in the prospection world, it is returned.

        :param obj: The object for which the corresponding object in the prospection World should be found.
        :return: The corresponding object in the prospection world.
        """
        with UseProspectionWorld():
            return self.world_sync.get_prospection_object(obj)

    def get_object_for_prospection_object(self, prospection_object: Object) -> Object:
        """
        Return the corresponding object from the main World for a given
        object in the prospection world. If the  given object is not in the prospection
        world an error will be raised.

        :param prospection_object: The object for which the corresponding object in the main World should be found.
        :return: The object in the main World.
        """
        with UseProspectionWorld():
            return self.world_sync.get_world_object(prospection_object)

    def remove_all_objects(self, exclude_objects: Optional[List[Object]] = None) -> None:
        """
        Remove all objects from the World.

        :param exclude_objects: A list of objects that should not be removed.
        """
        objs_copy = [obj for obj in self.objects]
        exclude_objects = [] if exclude_objects is None else exclude_objects
        [self.remove_object(obj) for obj in objs_copy if obj not in exclude_objects]

    def reset_world(self, remove_saved_states=False) -> None:
        """
        Reset the World to the state it was first spawned in.
        All attached objects will be detached, all joints will be set to the
        default position of 0 and all objects will be set to the position and
        orientation in which they were spawned.

        :param remove_saved_states: If the saved states should be removed.
        """
        self.restore_state(self.original_state_id)
        if remove_saved_states:
            self.remove_saved_states()
            self.original_state_id = self.save_state(use_same_id=True)

    def reset_concepts(self):
        """
        Reset the concepts of the World.
        """
        super().reset_concepts()
        [obj.reset_concepts() for obj in self.objects]

    def remove_saved_states(self) -> None:
        """
        Remove all saved states of the World.
        """
        if self.conf.use_physics_simulator_state:
            simulator_state_ids = set([state.simulator_state_id for state in self.saved_states.values()])
            for ssid in simulator_state_ids:
                self.remove_physics_simulator_state(ssid)
        else:
            self.remove_objects_saved_states()
        super().remove_saved_states()
        self.latest_state_id = None
        self.original_state_id = None

    def remove_objects_saved_states(self) -> None:
        """
        Remove all saved states of the objects in the World.
        """
        for obj in self.objects:
            obj.remove_saved_states()

    def update_transforms_for_objects_in_current_world(self) -> None:
        """
        Updates transformations for all objects that are currently in :py:attr:`~pycram.world.World.current_world`.
        """
        curr_time = Time().now()
        for obj in list(self.current_world.objects):
            obj.update_link_transforms(curr_time)

    def ray_test(self, from_position: List[float], to_position: List[float], calculate_distance: bool = False) \
            -> RayResult:
        """
        A wrapper around the :py:meth:`~pycram.world.World._ray_test` method that also calculates the distance
         of the ray if the calculate_distance parameter is set to True.

        :param from_position: The starting position of the ray in Cartesian world coordinates.
        :param to_position: The ending position of the ray in Cartesian world coordinates.
        :param calculate_distance: Whether to calculate the distance of the ray.
        :return: A RayResult object.
        """
        result = self._ray_test(from_position, to_position)
        if calculate_distance and not result.distance:
            result.update_distance(from_position, to_position)
        return result

    @abstractmethod
    def _ray_test(self, from_position: List[float], to_position: List[float]) -> RayResult:
        """ Cast a ray and return the first object hit, if any.

        :param from_position: The starting position of the ray in Cartesian world coordinates.
        :param to_position: The ending position of the ray in Cartesian world coordinates.
        :return: A RayResult object.
        """
        pass

    def ray_test_batch(self, from_positions: List[List[float]], to_positions: List[List[float]],
                       num_threads: int = 1, calculate_distances: bool = False) -> List[RayResult]:
        """
        A wrapper around the :py:meth:`~pycram.world.World._ray_test_batch` method that also calculates the distances
        of the rays if the calculate_distances parameter is set to True.

        :param from_positions: The starting positions of the rays in Cartesian world coordinates.
        :param to_positions: The ending positions of the rays in Cartesian world coordinates.
        :param num_threads: The number of threads to use to compute the ray intersections for the batch.
        :param calculate_distances: Whether to calculate the distances of the rays.
        :return: A list of RayResult objects.
        """
        results = self._ray_test_batch(from_positions, to_positions, num_threads)
        if calculate_distances:
            _ = [result.update_distance(from_positions[i], to_positions[i]) for i, result in enumerate(results)
                 if not result.distance]
        return results

    @abstractmethod
    def _ray_test_batch(self, from_positions: List[List[float]], to_positions: List[List[float]],
                        num_threads: int = 1) -> List[RayResult]:
        """ Cast a batch of rays and return the result for each of the rays (first object hit, if any. or -1)
         Takes optional argument num_threads to specify the number of threads to use
           to compute the ray intersections for the batch. Specify 0 to let simulator decide, 1 (default) for single
            core execution, 2 or more to select the number of threads to use.

        :param from_positions: The starting positions of the rays in Cartesian world coordinates.
        :param to_positions: The ending positions of the rays in Cartesian world coordinates.
        :param num_threads: The number of threads to use to compute the ray intersections for the batch.
        """
        pass

    def create_visual_shape(self, visual_shape: VisualShape) -> int:
        """
        Creates a visual shape in the physics simulator and returns the unique id of the created shape.

        :param visual_shape: The visual shape to be created, uses the VisualShape dataclass defined in world_dataclasses
        :return: The unique id of the created shape.
        """
        return self._simulator_object_creator(self._create_visual_shape, visual_shape)

    def _create_visual_shape(self, visual_shape: VisualShape) -> int:
        """
        See :py:meth:`~pycram.world.World.create_visual_shape`
        """
        raise NotImplementedError

    def create_multi_body_from_visual_shapes(self, visual_shape_ids: List[int], pose: PoseStamped) -> int:
        """
        Creates a multi body from visual shapes in the physics simulator and returns the unique id of the created
        multi body.

        :param visual_shape_ids: The ids of the visual shapes that should be used to create the multi body.
        :param pose: The pose of the origin of the multi body relative to the world frame.
        :return: The unique id of the created multi body.
        """
        # Dummy parameter since these are needed to spawn visual shapes as a multibody.
        num_of_shapes = len(visual_shape_ids)
        link_poses = [PoseStamped() for _ in range(num_of_shapes)]
        link_masses = [1.0 for _ in range(num_of_shapes)]
        link_parent = [0 for _ in range(num_of_shapes)]
        link_joints = [JointType.FIXED.value for _ in range(num_of_shapes)]
        link_collision = [-1 for _ in range(num_of_shapes)]
        link_joint_axis = [Point(x=1, y=0, z=0) for _ in range(num_of_shapes)]

        multi_body = MultiBody(base_visual_shape_index=-1, base_pose=pose,
                               link_visual_shape_indices=visual_shape_ids, link_poses=link_poses,
                               link_masses=link_masses,
                               link_inertial_frame_poses=link_poses,
                               link_parent_indices=link_parent, link_joint_types=link_joints,
                               link_joint_axis=link_joint_axis,
                               link_collision_shape_indices=link_collision)
        return self.create_multi_body(multi_body)

    def create_multi_body(self, multi_body: MultiBody) -> int:
        """
        Creates a multi body in the physics simulator and returns the unique id of the created multi body. The multibody
        is created by joining multiple links/shapes together with joints.

        :param multi_body: The multi body to be created, uses the MultiBody dataclass defined in world_dataclasses.
        :return: The unique id of the created multi body.
        """
        return self._simulator_object_creator(self._create_multi_body, multi_body)

    def _create_multi_body(self, multi_body: MultiBody) -> int:
        """
        See :py:meth:`~pycram.world.World.create_multi_body`
        """
        raise NotImplementedError

    def create_box_visual_shape(self, shape_data: BoxVisualShape) -> int:
        """
        Creates a box visual shape in the physics simulator and returns the unique id of the created shape.

        :param shape_data: The parameters that define the box visual shape to be created, uses the BoxVisualShape
         dataclass defined in world_dataclasses.
        :return: The unique id of the created shape.
        """
        return self._simulator_object_creator(self._create_box_visual_shape, shape_data)

    def _create_box_visual_shape(self, shape_data: BoxVisualShape) -> int:
        """
        See :py:meth:`~pycram.world.World.create_box_visual_shape`
        """
        raise NotImplementedError

    def create_cylinder_visual_shape(self, shape_data: CylinderVisualShape) -> int:
        """
        Creates a cylinder visual shape in the physics simulator and returns the unique id of the created shape.

        :param shape_data: The parameters that define the cylinder visual shape to be created, uses the
         CylinderVisualShape dataclass defined in world_dataclasses.
        :return: The unique id of the created shape.
        """
        return self._simulator_object_creator(self._create_cylinder_visual_shape, shape_data)

    def _create_cylinder_visual_shape(self, shape_data: CylinderVisualShape) -> int:
        """
        See :py:meth:`~pycram.world.World.create_cylinder_visual_shape`
        """
        raise NotImplementedError

    def create_sphere_visual_shape(self, shape_data: SphereVisualShape) -> int:
        """
        Creates a sphere visual shape in the physics simulator and returns the unique id of the created shape.

        :param shape_data: The parameters that define the sphere visual shape to be created, uses the SphereVisualShape
         dataclass defined in world_dataclasses.
        :return: The unique id of the created shape.
        """
        return self._simulator_object_creator(self._create_sphere_visual_shape, shape_data)

    def _create_sphere_visual_shape(self, shape_data: SphereVisualShape) -> int:
        """
        See :py:meth:`~pycram.world.World.create_sphere_visual_shape`
        """
        raise NotImplementedError

    def create_capsule_visual_shape(self, shape_data: CapsuleVisualShape) -> int:
        """
        Creates a capsule visual shape in the physics simulator and returns the unique id of the created shape.

        :param shape_data: The parameters that define the capsule visual shape to be created, uses the
         CapsuleVisualShape dataclass defined in world_dataclasses.
        :return: The unique id of the created shape.
        """
        return self._simulator_object_creator(self._create_capsule_visual_shape, shape_data)

    def _create_capsule_visual_shape(self, shape_data: CapsuleVisualShape) -> int:
        """
        See :py:meth:`~pycram.world.World.create_capsule_visual_shape`
        """
        raise NotImplementedError

    def create_plane_visual_shape(self, shape_data: PlaneVisualShape) -> int:
        """
        Creates a plane visual shape in the physics simulator and returns the unique id of the created shape.

        :param shape_data: The parameters that define the plane visual shape to be created, uses the PlaneVisualShape
         dataclass defined in world_dataclasses.
        :return: The unique id of the created shape.
        """
        return self._simulator_object_creator(self._create_plane_visual_shape, shape_data)

    def _create_plane_visual_shape(self, shape_data: PlaneVisualShape) -> int:
        """
        See :py:meth:`~pycram.world.World.create_plane_visual_shape`
        """
        raise NotImplementedError

    def create_mesh_visual_shape(self, shape_data: MeshVisualShape) -> int:
        """
        Creates a mesh visual shape in the physics simulator and returns the unique id of the created shape.

        :param shape_data: The parameters that define the mesh visual shape to be created,
        uses the MeshVisualShape dataclass defined in world_dataclasses.
        :return: The unique id of the created shape.
        """
        return self._simulator_object_creator(self._create_mesh_visual_shape, shape_data)

    def _create_mesh_visual_shape(self, shape_data: MeshVisualShape) -> int:
        """
        See :py:meth:`~pycram.world.World.create_mesh_visual_shape`
        """
        raise NotImplementedError

    def add_text(self, text: str, position: List[float], orientation: Optional[List[float]] = None, size: float = 0.1,
                 color: Optional[Color] = Color(), life_time: Optional[float] = 0,
                 parent_object_id: Optional[int] = None, parent_link_id: Optional[int] = None) -> int:
        """
        Adds text to the world.

        :param text: The text to be added.
        :param position: The position of the text in the world.
        :param orientation: By default, debug text will always face the camera, automatically rotation. By specifying a
         text orientation (quaternion), the orientation will be fixed in world space or local space
          (when parent is specified).
        :param size: The size of the text.
        :param color: The color of the text.
        :param life_time: The lifetime in seconds of the text to remain in the world, if 0 the text will remain in the
         world until it is removed manually.
        :param parent_object_id: The id of the object to which the text should be attached.
        :param parent_link_id: The id of the link to which the text should be attached.
        :return: The id of the added text.
        """
        return self._simulator_object_creator(self._add_text, text, position, orientation, size, color, life_time,
                                              parent_object_id, parent_link_id)

    def _add_text(self, text: str, position: List[float], orientation: Optional[List[float]] = None, size: float = 0.1,
                  color: Optional[Color] = Color(), life_time: Optional[float] = 0,
                  parent_object_id: Optional[int] = None, parent_link_id: Optional[int] = None) -> int:
        """
        See :py:meth:`~pycram.world.World.add_text`
        """
        raise NotImplementedError

    def remove_text(self, text_id: Optional[int] = None) -> None:
        """
        Removes text from the world using the given id. if no id is given all text will be removed.

        :param text_id: The id of the text to be removed.
        """
        self._simulator_object_remover(self._remove_text, text_id)

    def _remove_text(self, text_id: Optional[int] = None) -> None:
        """
        See :py:meth:`~pycram.world.World.remove_text`
        """
        raise NotImplementedError

    def enable_joint_force_torque_sensor(self, obj: Object, fts_joint_idx: int) -> None:
        """
        You can enable a joint force/torque sensor in each joint. Once enabled, if you perform
        a simulation step, the get_joint_reaction_force_torque will report the joint reaction forces in
        the fixed degrees of freedom: a fixed joint will measure all 6DOF joint forces/torques.
        A revolute/hinge joint force/torque sensor will measure 5DOF reaction forces along all axis except
        the hinge axis. The applied force by a joint motor is available through get_applied_joint_motor_torque.

        :param obj: The object in which the joint is located.
        :param fts_joint_idx: The index of the joint for which the force torque sensor should be enabled.
        """
        raise NotImplementedError

    def disable_joint_force_torque_sensor(self, obj: Object, joint_id: int) -> None:
        """
        Disables the force torque sensor of a joint.

        :param obj: The object in which the joint is located.
        :param joint_id: The id of the joint for which the force torque sensor should be disabled.
        """
        raise NotImplementedError

    def get_joint_reaction_force_torque(self, obj: Object, joint_id: int) -> List[float]:
        """
        Get the joint reaction forces and torques of the specified joint.

        :param obj: The object in which the joint is located.
        :param joint_id: The id of the joint for which the force torque should be returned.
        :return: The joint reaction forces and torques of the specified joint.
        """
        raise NotImplementedError

    def get_applied_joint_motor_torque(self, obj: Object, joint_id: int) -> float:
        """
        Get the applied torque by a joint motor.

        :param obj: The object in which the joint is located.
        :param joint_id: The id of the joint for which the applied motor torque should be returned.
        :return: The applied torque by a joint motor.
        """
        raise NotImplementedError

    def pause_world_sync(self) -> None:
        """
        Pause the world synchronization.
        """
        self.world_sync.sync_lock.acquire()

    def resume_world_sync(self) -> None:
        """
        Resume the world synchronization.
        """
        if self.world_sync.sync_lock.locked():
            self.world_sync.sync_lock.release()

    def add_vis_axis(self, pose: PoseStamped) -> int:
        """
        Add a visual axis to the world.

        :param pose: The pose of the visual axis.
        :return: The id of the added visual axis.
        """
        return self._simulator_object_creator(self._add_vis_axis, pose)

    def _add_vis_axis(self, pose: PoseStamped) -> None:
        """
        See :py:meth:`~pycram.world.World.add_vis_axis`
        """
        logwarn(f"Visual axis is not supported in {self.__class__.__name__}")

    def remove_vis_axis(self) -> None:
        """
        Remove the visual axis from the world.
        """
        self._simulator_object_remover(self._remove_vis_axis)

    def _remove_vis_axis(self) -> None:
        """
        See :py:meth:`~pycram.world.World.remove_vis_axis`
        """
        logwarn(f"Visual axis is not supported in {self.__class__.__name__}")

    def _simulator_object_creator(self, creator_func: Callable, *args, **kwargs) -> int:
        """
        Create an object in the physics simulator and returns the created object id.

        :param creator_func: The function that creates the object in the physics simulator.
        :param args: The arguments for the creator function.
        :param kwargs: The keyword arguments for the creator function.
        :return: The created object id.
        """
        obj_id = creator_func(*args, **kwargs)
        self.update_simulator_state_id_in_original_state()
        return obj_id

    def _simulator_object_remover(self, remover_func: Callable, *args, **kwargs) -> None:
        """
        Remove an object from the physics simulator.

        :param remover_func: The function that removes the object from the physics simulator.
        :param args: The arguments for the remover function.
        :param kwargs: The keyword arguments for the remover function.
        """
        remover_func(*args, **kwargs)
        self.update_simulator_state_id_in_original_state()

    def update_original_state(self) -> int:
        """
        Update the original state of the world.
        :return: The id of the updated original state.
        """
        return self.save_state(self.original_state_id, use_same_id=True)

    def update_simulator_state_id_in_original_state(self, use_same_id: bool = False) -> None:
        """
        Update the simulator state id in the original state if use_physics_simulator_state is True in the configuration.

        :param use_same_id: If the same id should be used for the state.
        """
        if self.conf.use_physics_simulator_state:
            self.original_state.simulator_state_id = self.save_physics_simulator_state(use_same_id=use_same_id)

    @property
    def original_state(self) -> WorldState:
        """
        The saved original state of the world.
        """
        return self.saved_states[self.original_state_id]

    def __eq__(self, other: World):
        if not isinstance(other, self.__class__):
            return False
        return (self.is_prospection_world == other.is_prospection_world
                and self.id == other.id)

    def __hash__(self):
        return hash((self.__class__.__name__, self.is_prospection_world, self.id))


class UseProspectionWorld:
    """
    An environment for using the prospection world, while in this environment the :py:attr:`~World.current_world`
    variable will point to the prospection world.

    Example:
        with UseProspectionWorld():
            NavigateAction.Action([[1, 0, 0], [0, 0, 0, 1]]).perform()
    """

    def __init__(self):
        self.prev_world: Optional[World] = None
        # The previous world is saved to restore it after the with block is exited.

    def __enter__(self):
        """
        This method is called when entering the with block, it will set the current world to the prospection world
        """
        # Please do not edit this function, it works as it is now!
        if not World.current_world.is_prospection_world:
            self.prev_world = World.current_world
            World.current_world = World.current_world.prospection_world
            # This is also a join statement since it is called from the main thread.
            World.current_world.world_sync.sync_worlds()

    def __exit__(self, *args):
        """
        This method is called when exiting the with block, it will restore the previous world to be the current world.
        """
        if self.prev_world is not None:
            World.current_world = self.prev_world


class WorldSync(threading.Thread):
    """
    Synchronizes the state between the World and its prospection world.
    Meaning the cartesian and joint position of everything in the prospection world will be
    synchronized with the main World.
    The class provides the possibility to pause the synchronization, this can be used
    if reasoning should be done in the prospection world.
    """

    WAIT_TIME_AS_N_SIMULATION_STEPS = 20
    """
    The time in simulation steps to wait between each iteration of the syncing loop.
    """

    def __init__(self, world: World, prospection_world: World):
        threading.Thread.__init__(self)
        self.world: World = world
        self.prospection_world: World = prospection_world
        self.prospection_world.world_sync = self

        self.terminate: bool = False
        self.pause_sync: bool = False
        # Maps world to prospection world objects
        self.object_to_prospection_object_map: Dict[Object, Object] = {}
        self.prospection_object_to_object_map: Dict[Object, Object] = {}
        self.equal_states = False
        self.sync_lock: threading.Lock = threading.Lock()

    def run(self):
        """
        Main method of the synchronization, this thread runs in a loop until the
        terminate flag is set.
        While this loop runs it continuously checks the cartesian and joint position of
        every object in the World and updates the corresponding object in the
        prospection world.
        """
        while not self.terminate:
            self.sync_lock.acquire()
            time.sleep(WorldSync.WAIT_TIME_AS_N_SIMULATION_STEPS * self.world.simulation_time_step)
            self.sync_lock.release()

    def get_world_object(self, prospection_object: Object) -> Object:
        """
        Get the corresponding object from the main World for a given object in the prospection world.

        :param prospection_object: The object for which the corresponding object in the main World should be found.
        :return: The object in the main World.
        """
        try:
            return self.prospection_object_to_object_map[prospection_object]
        except KeyError:
            if prospection_object in self.world.objects:
                return prospection_object
            raise ObjectNotFound(prospection_object)

    def get_prospection_object(self, obj: Object) -> Object:
        """
        Get the corresponding object from the prospection world for a given object in the main world.

        :param obj: The object for which the corresponding object in the prospection World should be found.
        :return: The corresponding object in the prospection world.
        """
        try:
            return self.object_to_prospection_object_map[obj]
        except KeyError:
            if obj in self.prospection_world.objects:
                return obj
            raise ProspectionObjectNotFound(obj)

    def sync_worlds(self):
        """
        Syncs the prospection world with the main world by adding and removing objects and synchronizing their states.
        """
        self.remove_objects_not_in_world()
        self.add_objects_not_in_prospection_world()
        self.prospection_object_to_object_map = {prospection_obj: obj for obj, prospection_obj in
                                                 self.object_to_prospection_object_map.items()}
        self.sync_objects_states()

    def remove_objects_not_in_world(self):
        """
        Removes all objects that are not in the main world from the prospection world.
        """
        obj_map_copy = copy(self.object_to_prospection_object_map)
        [self.remove_object(obj) for obj in obj_map_copy.keys() if obj not in self.world.objects]

    def add_objects_not_in_prospection_world(self):
        """
        Adds all objects that are in the main world but not in the prospection world to the prospection world.
        """
        obj_map_copy = copy(self.object_to_prospection_object_map)
        [self.add_object(obj) for obj in self.world.objects if obj not in obj_map_copy.keys()]

    def add_object(self, obj: Object) -> None:
        """
        Adds an object to the prospection world.

        :param obj: The object to be added.
        """
        self.object_to_prospection_object_map[obj] = obj.copy_to_prospection()

    def remove_object(self, obj: Object) -> None:
        """
        Removes an object from the prospection world.

        :param obj: The object to be removed.
        """
        prospection_obj = self.object_to_prospection_object_map[obj]
        prospection_obj.remove()
        del self.object_to_prospection_object_map[obj]

    def sync_objects_states(self) -> None:
        """
        Synchronizes the state of all objects in the World with the prospection world.
        """
        # Set the pose of the prospection objects to the pose of the world objects
        obj_pose_dict = {prospection_obj: obj.pose
                         for obj, prospection_obj in self.object_to_prospection_object_map.items()}
        for obj, prospection_obj in self.object_to_prospection_object_map.items():
            prospection_obj.set_attachments(obj.attachments)
            prospection_obj.joint_states = obj.joint_states
        self.world.prospection_world.reset_multiple_objects_base_poses(obj_pose_dict)

    def check_for_equal(self) -> bool:
        """
        Checks if both Worlds have the same state, meaning all objects are in the same position.
        This is currently not used, but might be used in the future if synchronization issues worsen.

        :return: True if both Worlds have the same state, False otherwise.
        """
        eql = True
        prospection_names = self.prospection_world.get_object_names()
        eql = eql and [name in prospection_names for name in self.world.get_object_names()]
        eql = eql and len(prospection_names) == len(self.world.get_object_names())
        if not eql:
            return False
        for obj, prospection_obj in self.object_to_prospection_object_map.items():
            eql = eql and obj.get_pose().position.euclidean_distance(prospection_obj.get_pose().position) < 0.001
        self.equal_states = eql
        return eql



from unittest import TestCase

import numpy as np
from pycram.tf_transformations import quaternion_from_euler

from pycram.datastructures.enums import JointType
from pycram.validation.error_checkers import calculate_angle_between_quaternions, \
    PoseErrorChecker, PositionErrorChecker, OrientationErrorChecker, RevoluteJointPositionErrorChecker, \
    PrismaticJointPositionErrorChecker, MultiJointPositionErrorChecker

from pycram.datastructures.pose import PoseStamped


class TestErrorCheckers(TestCase):
    @classmethod
    def setUpClass(cls):
        pass

    @classmethod
    def tearDownClass(cls):
        pass

    def tearDown(self):
        pass

    def test_calculate_quaternion_error(self):
        quat_1 = [0.0, 0.0, 0.0, 1.0]
        quat_2 = [0.0, 0.0, 0.0, 1.0]
        error = calculate_angle_between_quaternions(quat_1, quat_2)
        self.assertEqual(error, 0.0)
        quat_2 = quaternion_from_euler(0, 0, np.pi/2)
        error = calculate_angle_between_quaternions(quat_1, quat_2)
        self.assertEqual(error, np.pi/2)

    def test_pose_error_checker(self):
        pose_1 = PoseStamped.from_list([0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 1.0])
        pose_2 = PoseStamped.from_list([0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 1.0])
        error_checker = PoseErrorChecker()
        error = error_checker.calculate_error(pose_1, pose_2)
        self.assertEqual(error, [0.0, 0.0])
        self.assertTrue(error_checker.is_error_acceptable(pose_1, pose_2))
        quat = quaternion_from_euler(0, np.pi/2, 0)
        pose_2 = PoseStamped.from_list([0, 1, np.sqrt(3)], quat)
        error = error_checker.calculate_error(pose_1, pose_2)
        self.assertAlmostEqual(error[0], 2, places=5)
        self.assertEqual(error[1], np.pi/2)
        self.assertFalse(error_checker.is_error_acceptable(pose_1, pose_2))
        quat = quaternion_from_euler(0, 0, np.pi/360)
        pose_2 = PoseStamped.from_list([0, 0.0001, 0.0001], quat)
        self.assertTrue(error_checker.is_error_acceptable(pose_1, pose_2))
        quat = quaternion_from_euler(0, 0, np.pi / 179)
        pose_2 = PoseStamped.from_list([0, 0.0001, 0.0001], quat)
        self.assertFalse(error_checker.is_error_acceptable(pose_1, pose_2))

    def test_position_error_checker(self):
        position_1 = [0.0, 0.0, 0.0]
        position_2 = [0.0, 0.0, 0.0]
        error_checker = PositionErrorChecker()
        error = error_checker.calculate_error(position_1, position_2)
        self.assertEqual(error, 0.0)
        self.assertTrue(error_checker.is_error_acceptable(position_1, position_2))
        position_2 = [1.0, 1.0, 1.0]
        error = error_checker.calculate_error(position_1, position_2)
        self.assertAlmostEqual(error, np.sqrt(3), places=5)
        self.assertFalse(error_checker.is_error_acceptable(position_1, position_2))

    def test_orientation_error_checker(self):
        quat_1 = [0.0, 0.0, 0.0, 1.0]
        quat_2 = [0.0, 0.0, 0.0, 1.0]
        error_checker = OrientationErrorChecker()
        error = error_checker.calculate_error(quat_1, quat_2)
        self.assertEqual(error, 0.0)
        self.assertTrue(error_checker.is_error_acceptable(quat_1, quat_2))
        quat_2 = quaternion_from_euler(0, 0, np.pi/2)
        error = error_checker.calculate_error(quat_1, quat_2)
        self.assertEqual(error, np.pi/2)
        self.assertFalse(error_checker.is_error_acceptable(quat_1, quat_2))

    def test_revolute_joint_position_error_checker(self):
        position_1 = 0.0
        position_2 = 0.0
        error_checker = RevoluteJointPositionErrorChecker()
        error = error_checker.calculate_error(position_1, position_2)
        self.assertEqual(error, 0.0)
        self.assertTrue(error_checker.is_error_acceptable(position_1, position_2))
        position_2 = np.pi/2
        error = error_checker.calculate_error(position_1, position_2)
        self.assertEqual(error, np.pi/2)
        self.assertFalse(error_checker.is_error_acceptable(position_1, position_2))

    def test_prismatic_joint_position_error_checker(self):
        position_1 = 0.0
        position_2 = 0.0
        error_checker = PrismaticJointPositionErrorChecker()
        error = error_checker.calculate_error(position_1, position_2)
        self.assertEqual(error, 0.0)
        self.assertTrue(error_checker.is_error_acceptable(position_1, position_2))
        position_2 = 1.0
        error = error_checker.calculate_error(position_1, position_2)
        self.assertEqual(error, 1.0)
        self.assertFalse(error_checker.is_error_acceptable(position_1, position_2))

    def test_list_of_poses_error_checker(self):
        poses_1 = [PoseStamped.from_list([0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 1.0]),
                   PoseStamped.from_list([0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 1.0])]
        poses_2 = [PoseStamped.from_list([0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 1.0]),
                   PoseStamped.from_list([0.0, 0.0, 0.0], [0.0, 0.0, 0.0, 1.0])]
        error_checker = PoseErrorChecker(is_iterable=True)
        error = error_checker.calculate_error(poses_1, poses_2)
        self.assertEqual(error, [[0.0, 0.0], [0.0, 0.0]])
        self.assertTrue(error_checker.is_error_acceptable(poses_1, poses_2))
        quat = quaternion_from_euler(0, np.pi/2, 0)
        poses_2 = [PoseStamped.from_list([0, 1, np.sqrt(3)], quat),
                   PoseStamped.from_list([0, 1, np.sqrt(3)], quat)]
        error = error_checker.calculate_error(poses_1, poses_2)
        self.assertAlmostEqual(error[0][0], 2, places=5)
        self.assertEqual(error[0][1], np.pi/2)
        self.assertAlmostEqual(error[1][0], 2, places=5)
        self.assertEqual(error[1][1], np.pi/2)
        self.assertFalse(error_checker.is_error_acceptable(poses_1, poses_2))

    def test_multi_joint_error_checker(self):
        positions_1 = [0.0, 0.0]
        positions_2 = [np.pi/2, 0.1]
        joint_types = [JointType.REVOLUTE, JointType.PRISMATIC]
        error_checker = MultiJointPositionErrorChecker(joint_types)
        error = error_checker.calculate_error(positions_1, positions_2)
        self.assertEqual(error, [np.pi/2, 0.1])
        self.assertFalse(error_checker.is_error_acceptable(positions_1, positions_2))
        positions_2 = [np.pi/180, 0.0001]
        self.assertTrue(error_checker.is_error_acceptable(positions_1, positions_2))



from __future__ import annotations

import glob
import os
import pathlib
import shutil

from typing_extensions import List, TYPE_CHECKING, Optional

from .ros import  loginfo

if TYPE_CHECKING:
    from .description import ObjectDescription
    from .datastructures.dataclasses import Color
    from .datastructures.pose import TransformStamped


class CacheManager:

    """
    The CacheManager is responsible for caching object description files and managing the cache directory.
    """

    cache_cleared: bool = False
    """
    Indicate whether the cache directory has been cleared at least once since beginning or not.
    """

    def __init__(self, cache_dir: str, data_directory: List[str], clear_cache: bool = True):
        """
        Initialize the CacheManager.

        :param cache_dir: The directory where the cached files are stored.
        :param data_directory: The directory where all resource files are stored.
        :param clear_cache: If True, the cache directory will be cleared.
        """
        self.cache_dir = cache_dir
        self.data_directories = data_directory
        if clear_cache:
            self.clear_cache()

    def clear_cache(self):
        """
        Clear the cache directory.
        """
        self.delete_cache_dir()
        self.create_cache_dir_if_not_exists()
        self.cache_cleared = True

    def delete_cache_dir(self):
        """
        Delete the cache directory.
        """
        if pathlib.Path(self.cache_dir).exists():
            shutil.rmtree(self.cache_dir)

    def update_cache_dir_with_object(self, path: str, ignore_cached_files: bool,
                                     object_description: ObjectDescription, object_name: str,
                                     scale_mesh: Optional[float] = None,
                                     mesh_transform: Optional[TransformStamped] = None,
                                     color: Optional[Color] = None) -> str:
        """
        Check if the file is already in the cache directory, if not preprocess and save in the cache.

        :param path: The path of the file to preprocess and save in the cache directory.
        :param ignore_cached_files: If True, the file will be preprocessed and saved in the cache directory even if it
         is already cached.
        :param object_description: The object description of the file.
        :param object_name: The name of the object.
        :param scale_mesh: The scale of the mesh.
        :param mesh_transform: The transformation matrix to apply to the mesh.
        :param color: The color of the object.
        :return: The path of the cached file.
        """
        path_object = pathlib.Path(path)
        extension = path_object.suffix

        self.create_cache_dir_if_not_exists()

        # save correct path in case the file is already in the cache directory
        cache_path = os.path.join(self.cache_dir, object_description.get_file_name(path_object, extension, object_name))

        if not self.is_cached(path, object_description) or ignore_cached_files:
            # if file is not yet cached preprocess the description file and save it in the cache directory.
            path = self.look_for_file_in_data_dir(path_object)
            object_description.original_path = path
            object_description.generate_description_from_file(path, object_name, extension, cache_path,
                                                              scale_mesh, mesh_transform, color)

        return cache_path

    def look_for_file_in_data_dir(self, path_object: pathlib.Path) -> str:
        """
        Look for a file in the data directory of the World. If the file is not found in the data directory, raise a
         FileNotFoundError.

        :param path_object: The pathlib object of the file to look for.
        """
        name = path_object.name
        for data_dir in self.data_directories:
            data_path = pathlib.Path(data_dir).joinpath("**")
            for file in glob.glob(str(data_path), recursive=True):
                file_path = pathlib.Path(file)
                if file_path.name == name:
                    loginfo(f"Found file {name} in {file_path}")
                    return str(file_path)

        raise FileNotFoundError(
            f"File {name} could not be found in the resource directory {self.data_directories}")

    def create_cache_dir_if_not_exists(self):
        """
        Create the cache directory if it does not exist.
        """
        if not pathlib.Path(self.cache_dir).exists():
            os.mkdir(self.cache_dir)

    def is_cached(self, path: str, object_description: ObjectDescription) -> bool:
        """
        Check if the file in the given path is already cached or if
        there is already a cached file with the given name, this is the case if a .stl, .obj file or a description from
        the parameter server is used.

        :param path: The path of the file to check.
        :param object_description: The object description of the file.
        :return: True if there already exists a cached file, False in any other case.
        """
        return self.check_with_extension(path) or self.check_without_extension(path, object_description)

    def check_with_extension(self, path: str) -> bool:
        """
        Check if the file in the given ath exists in the cache directory including file extension.

        :param path: The path of the file to check.
        """
        file_name = pathlib.Path(path).name
        full_path = pathlib.Path(os.path.join(self.cache_dir, file_name))
        return full_path.exists()

    def check_without_extension(self, path: str, object_description: ObjectDescription) -> bool:
        """
        Check if the file in the given path exists in the cache directory the given file extension.
        Instead, replace the given extension with the extension of the used ObjectDescription and check for that one.

        :param path: The path of the file to check.
        :param object_description: The object description of the file.
        """
        file_stem = pathlib.Path(path).stem
        full_path = pathlib.Path(os.path.join(self.cache_dir, file_stem + object_description.get_file_extension()))
        return full_path.exists()



# used for delayed evaluation of typing until python 3.11 becomes mainstream
from __future__ import annotations

import abc
import inspect
import math
from dataclasses import dataclass, field
from datetime import timedelta
from functools import cached_property
from time import sleep

import numpy as np

from .object_designator import BelieveObject
from ..datastructures.world_entity import PhysicalBody
from ..has_parameters import has_parameters
from ..language import SequentialPlan, TryInOrderPlan
from ..plan import with_plan

from ..datastructures.partial_designator import PartialDesignator
from ..datastructures.dataclasses import FrozenObject

from .. import utils
from ..tf_transformations import quaternion_from_euler
from typing_extensions import List, Union, Optional, Type, Dict, Any, Iterable

from pycrap.ontologies import Location, PhysicalObject
from .location_designator import CostmapLocation
from .motion_designator import MoveJointsMotion, MoveGripperMotion, MoveTCPMotion, MoveMotion, \
    LookingMotion, DetectingMotion, OpeningMotion, ClosingMotion
from ..datastructures.grasp import GraspDescription
from ..datastructures.world import World, UseProspectionWorld
from ..description import Joint, Link, ObjectDescription
from ..designator import ActionDescription, ObjectDesignatorDescription
from ..failure_handling import try_action
from ..failures import TorsoGoalNotReached, ConfigurationNotReached, ObjectNotInGraspingArea, \
    ObjectNotPlacedAtTargetLocation, ObjectStillInContact, LookAtGoalNotReached, \
    ContainerManipulationError
from ..local_transformer import LocalTransformer
from ..failures import ObjectUnfetchable, ReachabilityFailure, NavigationGoalNotReachedError, PerceptionObjectNotFound, \
    ObjectNotGraspedError
from ..robot_description import EndEffectorDescription
from ..ros import sleep
from ..config.action_conf import ActionConfig

from ..datastructures.enums import Arms, Grasp, GripperState, DetectionTechnique, DetectionState, MovementType, \
    TorsoState, StaticJointState, Frame, FindBodyInRegionMethod, ContainerManipulationType

from ..datastructures.pose import PoseStamped
from ..datastructures.world import World

from ..robot_description import RobotDescription, KinematicChainDescription
from ..ros import logwarn, loginfo
from ..validation.error_checkers import PoseErrorChecker
from ..validation.goal_validator import create_multiple_joint_goal_validator
from ..world_concepts.world_object import Object
from ..world_reasoning import move_away_all_objects_to_create_empty_space, generate_object_at_target, \
    cast_a_ray_from_camera, has_gripper_grasped_body, is_body_between_fingers


# ----------------------------------------------------------------------------
# ---------------- Performables ----------------------------------------------
# ----------------------------------------------------------------------------

def record_object_pre_perform(action):
    """
    Record the object before the action is performed.

    This should be appended to the pre performs of every action that interacts with an object.
    """
    # for every field in the action that is an object
    # write it to a dict mapping the OG field name to the frozen copy
    action.object_at_execution = action.object_designator.frozen_copy()


@has_parameters
@dataclass
class MoveTorsoAction(ActionDescription):
    """
    Move the torso of the robot up and down.
    """
    torso_state: TorsoState
    """
    The state of the torso that should be set
    """

    def plan(self) -> None:
        joint_positions: dict = RobotDescription.current_robot_description.get_static_joint_chain("torso",
                                                                                                  self.torso_state)
        MoveJointsMotion(list(joint_positions.keys()), list(joint_positions.values())).perform()

    def validate(self, result: Optional[Any] = None, max_wait_time: timedelta = timedelta(seconds=2)):
        """
        Create a goal validator for the joint positions and wait until the goal is achieved or the timeout is reached.
        """

        joint_positions: dict = RobotDescription.current_robot_description.get_static_joint_chain("torso",
                                                                                                  self.torso_state)
        validator = create_multiple_joint_goal_validator(World.current_world.robot, joint_positions)
        validator.wait_until_goal_is_achieved(max_wait_time=max_wait_time,
                                              time_per_read=timedelta(milliseconds=20))
        if not validator.goal_achieved:
            raise TorsoGoalNotReached(validator)

    @classmethod
    @with_plan
    def description(cls, torso_state: Union[Iterable[TorsoState], TorsoState]) -> PartialDesignator[
        Type[MoveTorsoAction]]:
        return PartialDesignator(MoveTorsoAction, torso_state=torso_state)


@has_parameters
@dataclass
class SetGripperAction(ActionDescription):
    """
    Set the gripper state of the robot.
    """

    gripper: Arms
    """
    The gripper that should be set 
    """
    motion: GripperState
    """
    The motion that should be set on the gripper
    """

    def plan(self) -> None:
        arm_chains = RobotDescription.current_robot_description.get_arm_chain(self.gripper)
        if type(arm_chains) is not list:
            MoveGripperMotion(gripper=arm_chains.arm_type, motion=self.motion).perform()
        else:
            for chain in arm_chains:
                MoveGripperMotion(gripper=chain.arm_type, motion=self.motion).perform()

    def validate(self, result: Optional[Any] = None, max_wait_time: timedelta = timedelta(seconds=2)):
        """
        Needs gripper state to be read or perceived.
        """
        pass

    @classmethod
    @with_plan
    def description(cls, gripper: Union[Iterable[Arms], Arms],
                    motion: Union[Iterable[GripperState], GripperState] = None) -> PartialDesignator[
        Type[SetGripperAction]]:
        return PartialDesignator(SetGripperAction, gripper=gripper, motion=motion)


@has_parameters
@dataclass
class ReleaseAction(ActionDescription):
    """
    Releases an Object from the robot.

    Note: This action can not ve used yet.
    """
    object_designator: Object
    """
    The object that should be released
    """

    object_at_execution: Optional[FrozenObject] = field(init=False, repr=False, default=None)
    """
    The object at the time this Action got created. It is used to be a static, information holding entity
    """

    gripper: Arms
    """
    The gripper that should be used to release the object
    """

    _pre_perform_callbacks = []
    """
    List to save the callbacks which should be called before performing the action.
    """

    def __post_init__(self):
        super().__post_init__()

        # Store the object's data copy at execution
        self.pre_perform(record_object_pre_perform)

    def plan(self) -> None:
        raise NotImplementedError

    @classmethod
    def description(cls, object_designator: ObjectDesignatorDescription,
                    gripper: Optional[Union[Iterable[Arms], Arms]] = None) -> PartialDesignator[
        Type[ReleaseAction]]:
        return PartialDesignator(ReleaseAction, object_designator=object_designator, gripper=gripper)


@has_parameters
@dataclass
class GripAction(ActionDescription):
    """
    Grip an object with the robot.

    Note: This action can not be used yet.
    """
    object_designator: Object
    """
    The object that should be gripped
    """

    object_at_execution: Optional[FrozenObject] = field(init=False, repr=False, default=None)
    """
    The object at the time this Action got created. It is used to be a static, information holding entity
    """

    gripper: Arms
    """
    The gripper that should be used to grip the object
    """

    effort: float = None
    """
    The effort that should be used to grip the object
    """

    _pre_perform_callbacks = []
    """
    List to save the callbacks which should be called before performing the action.
    """

    def __post_init__(self):
        super().__post_init__()

        # Store the object's data copy at execution
        self.pre_perform(record_object_pre_perform)

    def plan(self) -> None:
        raise NotImplementedError()

    @classmethod
    @with_plan
    def description(cls, object_designator: Union[Iterable[Object], Object],
                    gripper: Union[Iterable[Arms], Arms] = None, effort: Union[Iterable[float], float] = None, ) -> \
            PartialDesignator[Type[GripAction]]:
        return PartialDesignator(GripAction, object_designator=object_designator,
                                 gripper=gripper, effort=effort)


@has_parameters
@dataclass
class ParkArmsAction(ActionDescription):
    """
    Park the arms of the robot.
    """

    arm: Arms
    """
    Entry from the enum for which arm should be parked
    """

    def plan(self) -> None:
        joint_poses = self.get_joint_poses()
        MoveJointsMotion(names=list(joint_poses.keys()), positions=list(joint_poses.values())).perform()

    def get_joint_poses(self) -> Dict[str, float]:
        """
        :return: The joint positions that should be set for the arm to be in the park position.
        """
        joint_poses = {}
        arm_chains = RobotDescription.current_robot_description.get_arm_chain(self.arm)
        if type(arm_chains) is not list:
            joint_poses = arm_chains.get_static_joint_states(StaticJointState.Park)
        else:
            for arm_chain in RobotDescription.current_robot_description.get_arm_chain(self.arm):
                joint_poses.update(arm_chain.get_static_joint_states(StaticJointState.Park))
        return joint_poses

    def validate(self, result: Optional[Any] = None, max_wait_time: timedelta = timedelta(seconds=2)):
        """
        Create a goal validator for the joint positions and wait until the goal is achieved or the timeout is reached.
        """
        joint_poses = self.get_joint_poses()
        validator = create_multiple_joint_goal_validator(World.current_world.robot, joint_poses)
        validator.wait_until_goal_is_achieved(max_wait_time=max_wait_time,
                                              time_per_read=timedelta(milliseconds=20))
        if not validator.goal_achieved:
            raise ConfigurationNotReached(validator, configuration_type=StaticJointState.Park)

    @classmethod
    @with_plan
    def description(cls, arm: Union[Iterable[Arms], Arms]) -> PartialDesignator[Type[ParkArmsAction]]:
        return PartialDesignator(ParkArmsAction, arm=arm)


SPECIAL_KNOWLEDGE = {
    'bigknife':
        [("top", [-0.08, 0, 0])],
    'whisk':
        [("top", [-0.08, 0, 0])],
    'bowl':
        [("front", [1.0, 2.0, 3.0]),
         ("key2", [4.0, 5.0, 6.0])]
}


@has_parameters
@dataclass
class ReachToPickUpAction(ActionDescription):
    """
    Let the robot reach a specific pose.
    """

    object_designator: Object
    """
    Object designator_description describing the object that should be picked up
    """

    arm: Arms
    """
    The arm that should be used for pick up
    """

    grasp_description: GraspDescription
    """
    The grasp description that should be used for picking up the object
    """

    object_at_execution: Optional[FrozenObject] = field(init=False, repr=False, default=None)
    """
    The object at the time this Action got created. It is used to be a static, information holding entity. It is
    not updated when the world object is changed.
    """
    _pre_perform_callbacks = []
    """
    List to save the callbacks which should be called before performing the action.
    """

    def __post_init__(self):
        super().__post_init__()

        # Store the object's data copy at execution
        self.pre_perform(record_object_pre_perform)

    def plan(self) -> None:

        target_pose = self.object_designator.get_grasp_pose(self.end_effector, self.grasp_description)
        target_pose.rotate_by_quaternion(self.end_effector.grasps[self.grasp_description])

        target_pre_pose = LocalTransformer().translate_pose_along_local_axis(target_pose,
                                                                             self.end_effector.get_approach_axis(),
                                                                             -self.object_designator.get_approach_offset())

        MoveGripperMotion(motion=GripperState.OPEN, gripper=self.arm).perform()

        self.move_gripper_to_pose(target_pre_pose)

        self.move_gripper_to_pose(target_pose, MovementType.STRAIGHT_CARTESIAN)

        # Remove the vis axis from the world if it was added
        World.current_world.remove_vis_axis()

    def move_gripper_to_pose(self, pose: PoseStamped, movement_type: MovementType = MovementType.CARTESIAN,
                             add_vis_axis: bool = True):
        """
        Move the gripper to a specific pose.

        :param pose: The pose to go to.
        :param movement_type: The type of movement that should be performed.
        :param add_vis_axis: If a visual axis should be added to the world.
        """
        pose = self.local_transformer.transform_pose(pose, Frame.Map.value)
        if add_vis_axis:
            World.current_world.add_vis_axis(pose)
        MoveTCPMotion(pose, self.arm, allow_gripper_collision=False, movement_type=movement_type).perform()

    @cached_property
    def local_transformer(self) -> LocalTransformer:
        return LocalTransformer()

    @cached_property
    def arm_chain(self) -> KinematicChainDescription:
        return RobotDescription.current_robot_description.get_arm_chain(self.arm)

    @cached_property
    def end_effector(self) -> EndEffectorDescription:
        return self.arm_chain.end_effector

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        """
        Check if object is contained in the gripper such that it can be grasped and picked up.
        """
        fingers_link_names = self.arm_chain.end_effector.fingers_link_names
        if fingers_link_names:
            if not is_body_between_fingers(self.object_designator, fingers_link_names,
                                           method=FindBodyInRegionMethod.MultiRay):
                raise ObjectNotInGraspingArea(self.object_designator, World.robot, self.arm, self.grasp_description)
        else:
            logwarn(f"Cannot validate reaching to pick up action for arm {self.arm} as no finger links are defined.")

    @classmethod
    @with_plan
    def description(cls, object_designator: Union[Iterable[Object], Object],
                    arm: Union[Iterable[Arms], Arms] = None,
                    grasp: Union[Iterable[Grasp], Grasp] = None) -> PartialDesignator[Type[ReachToPickUpAction]]:
        return PartialDesignator(ReachToPickUpAction, object_designator=object_designator,
                                 arm=arm,
                                 grasp=grasp)


@has_parameters
@dataclass
class PickUpAction(ActionDescription):
    """
    Let the robot pick up an object.
    """

    object_designator: Object
    """
    Object designator_description describing the object that should be picked up
    """

    arm: Arms
    """
    The arm that should be used for pick up
    """

    grasp_description: GraspDescription
    """
    The GraspDescription that should be used for picking up the object
    """

    object_at_execution: Optional[FrozenObject] = field(init=False, repr=False, default=None)
    """
    The object at the time this Action got created. It is used to be a static, information holding entity. It is
    not updated when the BulletWorld object is changed.
    """

    _pre_perform_callbacks = []
    """
    List to save the callbacks which should be called before performing the action.
    """

    def __post_init__(self):
        super().__post_init__()

        # Store the object's data copy at execution
        self.pre_perform(record_object_pre_perform)

    def plan(self) -> None:
        ReachToPickUpAction(self.object_designator, self.arm, self.grasp_description).perform()

        MoveGripperMotion(motion=GripperState.CLOSE, gripper=self.arm).perform()

        tool_frame = RobotDescription.current_robot_description.get_arm_chain(self.arm).get_tool_frame()
        World.robot.attach(self.object_designator, tool_frame)

        self.lift_object(distance=0.1)

        # Remove the vis axis from the world
        World.current_world.remove_vis_axis()

    def lift_object(self, distance: float = 0.1):
        lift_to_pose = self.gripper_pose()
        lift_to_pose.pose.position.z += distance
        MoveTCPMotion(lift_to_pose, self.arm, allow_gripper_collision=True).perform()

    def gripper_pose(self) -> PoseStamped:
        """
        Get the pose of the gripper.

        :return: The pose of the gripper.
        """
        gripper_link = self.arm_chain.get_tool_frame()
        return World.robot.links[gripper_link].pose

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        """
        Check if picked up object is in contact with the gripper.
        """
        if not has_gripper_grasped_body(self.arm, self.object_designator):
            raise ObjectNotGraspedError(self.object_designator, World.robot, self.arm, self.grasp_description)

    @cached_property
    def arm_chain(self) -> KinematicChainDescription:
        return RobotDescription.current_robot_description.get_arm_chain(self.arm)

    @classmethod
    @with_plan
    def description(cls, object_designator: Union[Iterable[Object], Object],
                    arm: Union[Iterable[Arms], Arms] = None,
                    grasp_description: Union[Iterable[GraspDescription], GraspDescription] = None) -> \
            PartialDesignator[Type[PickUpAction]]:
        return PartialDesignator(PickUpAction, object_designator=object_designator, arm=arm,
                                 grasp_description=grasp_description)


@has_parameters
@dataclass
class PlaceAction(ActionDescription):
    """
    Places an Object at a position using an arm.
    """

    object_designator: Object
    """
    Object designator_description describing the object that should be place
    """
    target_location: PoseStamped
    """
    Pose in the world at which the object should be placed
    """
    arm: Arms
    """
    Arm that is currently holding the object
    """
    object_at_execution: Optional[FrozenObject] = field(init=False, repr=False, default=None)
    """
    The object at the time this Action got created. It is used to be a static, information holding entity. It is
    not updated when the BulletWorld object is changed.
    """
    _pre_perform_callbacks = []
    """
    List to save the callbacks which should be called before performing the action.
    """

    def __post_init__(self):
        super().__post_init__()

        # Store the object's data copy at execution
        self.pre_perform(record_object_pre_perform)

    def plan(self) -> None:
        target_pose = self.object_designator.attachments[
            World.robot].get_child_link_target_pose_given_parent(self.target_location)
        MoveTCPMotion(target_pose, self.arm).perform()

        MoveGripperMotion(GripperState.OPEN, self.arm).perform()
        World.robot.detach(self.object_designator)

        retract_pose = LocalTransformer().translate_pose_along_local_axis(target_pose,
                                                                          self.end_effector.get_approach_axis(),
                                                                          -self.object_designator.get_approach_offset())
        MoveTCPMotion(retract_pose, self.arm).perform()

    @cached_property
    def gripper_link(self) -> Link:
        return World.robot.links[self.arm_chain.get_tool_frame()]

    @cached_property
    def arm_chain(self) -> KinematicChainDescription:
        return RobotDescription.current_robot_description.get_arm_chain(self.arm)

    @cached_property
    def end_effector(self) -> EndEffectorDescription:
        return self.arm_chain.end_effector

    @cached_property
    def local_transformer(self) -> LocalTransformer:
        return LocalTransformer()

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        """
        Check if the object is placed at the target location.
        """
        self.validate_loss_of_contact()
        self.validate_placement_location()

    def validate_loss_of_contact(self):
        """
        Check if the object is still in contact with the robot after placing it.
        """
        contact_links = self.object_designator.get_contact_points_with_body(World.robot).get_all_bodies()
        if contact_links:
            raise ObjectStillInContact(self.object_designator, contact_links,
                                       self.target_location, World.robot, self.arm)

    def validate_placement_location(self):
        """
        Check if the object is placed at the target location.
        """
        pose_error_checker = PoseErrorChecker(World.conf.get_pose_tolerance())
        if not pose_error_checker.is_error_acceptable(self.object_designator.pose, self.target_location):
            raise ObjectNotPlacedAtTargetLocation(self.object_designator, self.target_location, World.robot, self.arm)

    @classmethod
    @with_plan
    def description(cls, object_designator: Union[Iterable[Object], Object],
                    target_location: Union[Iterable[PoseStamped], PoseStamped],
                    arm: Union[Iterable[Arms], Arms] = None) -> PartialDesignator[Type[PlaceAction]]:
        return PartialDesignator(PlaceAction, object_designator=object_designator,
                                 target_location=target_location,
                                 arm=arm)


@has_parameters
@dataclass
class NavigateAction(ActionDescription):
    """
    Navigates the Robot to a position.
    """

    target_location: PoseStamped
    """
    Location to which the robot should be navigated
    """

    keep_joint_states: bool = ActionConfig.navigate_keep_joint_states
    """
    Keep the joint states of the robot the same during the navigation.
    """

    def plan(self) -> None:
        motion_action = MoveMotion(self.target_location, self.keep_joint_states)
        return try_action(motion_action, failure_type=NavigationGoalNotReachedError)

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        pose_validator = PoseErrorChecker(World.conf.get_pose_tolerance())
        if not pose_validator.is_error_acceptable(World.robot.pose, self.target_location):
            raise NavigationGoalNotReachedError(World.robot.pose, self.target_location)

    @classmethod
    @with_plan
    def description(cls, target_location: Union[Iterable[PoseStamped], PoseStamped],
                    keep_joint_states: Union[Iterable[bool], bool] = ActionConfig.navigate_keep_joint_states) -> \
            PartialDesignator[Type[NavigateAction]]:
        return PartialDesignator(NavigateAction, target_location=target_location,
                                 keep_joint_states=keep_joint_states)


@has_parameters
@dataclass
class TransportAction(ActionDescription):
    """
    Transports an object to a position using an arm
    """

    object_designator: Object
    """
    Object designator_description describing the object that should be transported.
    """
    target_location: PoseStamped
    """
    Target Location to which the object should be transported
    """
    arm: Arms
    """
    Arm that should be used
    """
    _pre_perform_callbacks = []
    """
    List to save the callbacks which should be called before performing the action.
    """

    def __post_init__(self):
        super().__post_init__()

        # Store the object's data copy at execution
        self.pre_perform(record_object_pre_perform)

    def plan(self) -> None:
        robot_desig_resolved = BelieveObject(names=[RobotDescription.current_robot_description.name]).resolve()
        ParkArmsActionDescription(Arms.BOTH).perform()
        pickup_loc = CostmapLocation(target=self.object_designator,
                                     reachable_for=robot_desig_resolved,
                                     reachable_arm=[self.arm])
        # Tries to find a pick-up position for the robot that uses the given arm
        pickup_pose = pickup_loc.resolve()
        if not pickup_pose:
            raise ObjectUnfetchable(
                f"Found no pose for the robot to grasp the object: {self.object_designator} with arm: {self.arm}")

        NavigateActionDescription(pickup_pose, True).perform()
        PickUpActionDescription(self.object_designator, pickup_pose.arm,
                     grasp_description=pickup_pose.grasp_description).perform()
        ParkArmsActionDescription(Arms.BOTH).perform()
        try:
            place_loc = CostmapLocation(
                target=self.target_location,
                reachable_for=robot_desig_resolved,
                reachable_arm=[pickup_pose.arm],
                grasp_descriptions=[pickup_pose.grasp_description],
                object_in_hand=self.object_designator
            ).resolve()
        except StopIteration:
            raise ReachabilityFailure(
                f"No location found from where the robot can reach the target location: {self.target_location}")
        NavigateActionDescription(place_loc, True).perform()
        PlaceActionDescription(self.object_designator, self.target_location, self.arm).perform()
        ParkArmsActionDescription(Arms.BOTH).perform()

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        # The validation of each atomic action is done in the action itself, so no more validation needed here.
        pass

    @classmethod
    @with_plan
    def description(cls, object_designator: Union[Iterable[Object], Object],
                    target_location: Union[Iterable[PoseStamped], PoseStamped],
                    arm: Union[Iterable[Arms], Arms] = None) -> PartialDesignator[Type[TransportAction]]:
        return PartialDesignator(TransportAction, object_designator=object_designator,
                                 target_location=target_location,
                                 arm=arm)


@has_parameters
@dataclass
class LookAtAction(ActionDescription):
    """
    Lets the robot look at a position.
    """

    target: PoseStamped
    """
    Position at which the robot should look, given as 6D pose
    """

    def plan(self) -> None:
        LookingMotion(target=self.target).perform()

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        """
        Check if the robot is looking at the target location by spawning a virtual object at the target location and
        creating a ray from the camera and checking if it intersects with the object.
        """
        return
        with UseProspectionWorld():
            move_away_all_objects_to_create_empty_space(exclude_objects=[World.robot.name, "floor"])
            # Create a virtual object at the target location, the current size is 40x40x40 cm which is very big in
            # my opinion, maybe this indicates that the looking at action is not accurate # TODO check this
            gen_obj = generate_object_at_target(self.target.position.to_list(), size=(0.4, 0.4, 0.4))
            ray_result = cast_a_ray_from_camera()
            gen_obj.remove()
            if not ray_result.intersected or ray_result.obj_id != gen_obj.id:
                raise LookAtGoalNotReached(World.robot, self.target)

    @classmethod
    @with_plan
    def description(cls, target: Union[Iterable[PoseStamped], PoseStamped]) -> PartialDesignator[Type[LookAtAction]]:
        return PartialDesignator(LookAtAction, target=target)


@has_parameters
@dataclass
class DetectAction(ActionDescription):
    """
    Detects an object that fits the object description and returns an object designator_description describing the object.

    If no object is found, an PerceptionObjectNotFound error is raised.
    """

    technique: DetectionTechnique
    """
    The technique that should be used for detection
    """
    state: DetectionState = None
    """
    The state of the detection, e.g Start Stop for continues perception
    """
    object_designator: Optional[Object] = None
    """
    The type of the object that should be detected, only considered if technique is equal to Type
    """
    region: Location = None
    """
    The region in which the object should be detected
    """

    # object_at_execution: Optional[FrozenObject] = field(init=False, repr=False, default=None)
    # """
    # The object at the time this Action got created. It is used to be a static, information holding entity
    # """

    _pre_perform_callbacks = []
    """
    List to save the callbacks which should be called before performing the action.
    """

    def __post_init__(self):
        super().__post_init__()

        # # Store the object's data copy at execution
        # self.pre_perform(record_object_pre_perform)

    def plan(self) -> None:
        return try_action(DetectingMotion(technique=self.technique, state=self.state,
                                          object_designator_description=self.object_designator,
                                          region=self.region), PerceptionObjectNotFound)

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        return
        # if not result:
        #     raise PerceptionObjectNotFound(self.object_designator, self.technique, self.region)

    @classmethod
    @with_plan
    def description(cls, technique: Union[Iterable[DetectionTechnique], DetectionTechnique],
                    state: Union[Iterable[DetectionState], DetectionState] = None,
                    object_designator: Union[Iterable[Object], Object] = None,
                    region: Union[Iterable[Location], Location] = None) -> PartialDesignator[Type[DetectAction]]:
        return PartialDesignator(DetectAction, technique=technique,
                                 state=state,
                                 object_designator=object_designator,
                                 region=region)


@has_parameters
@dataclass
class OpenAction(ActionDescription):
    """
    Opens a container like object
    """

    object_designator: ObjectDescription.Link
    """
    Object designator_description describing the object that should be opened
    """
    arm: Arms
    """
    Arm that should be used for opening the container
    """
    grasping_prepose_distance: float = ActionConfig.grasping_prepose_distance
    """
    The distance in meters the gripper should be at in the x-axis away from the handle.
    """

    def plan(self) -> None:
        GraspingAction(self.object_designator, self.arm, self.grasping_prepose_distance).perform()
        OpeningMotion(self.object_designator, self.arm).perform()

        MoveGripperMotion(GripperState.OPEN, self.arm, allow_gripper_collision=True).perform()

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        """
        Check if the container is opened, this assumes that the container state can be read accurately from the
        real world.
        """
        validate_close_open(self.object_designator, self.arm, OpenAction)

    @classmethod
    @with_plan
    def description(cls, object_designator_description: Union[Iterable[ObjectDescription.Link], ObjectDescription.Link],
                    arm: Union[Iterable[Arms], Arms] = None,
                    grasping_prepose_distance: Union[
                        Iterable[float], float] = ActionConfig.grasping_prepose_distance) -> \
            PartialDesignator[Type[OpenAction]]:
        return PartialDesignator(OpenAction, object_designator=object_designator_description,
                                 arm=arm,
                                 grasping_prepose_distance=grasping_prepose_distance)


@has_parameters
@dataclass
class CloseAction(ActionDescription):
    """
    Closes a container like object.
    """

    object_designator: ObjectDescription.Link
    """
    Object designator_description describing the object that should be closed
    """
    arm: Arms
    """
    Arm that should be used for closing
    """
    grasping_prepose_distance: float = ActionConfig.grasping_prepose_distance
    """
    The distance in meters between the gripper and the handle before approaching to grasp.
    """

    def plan(self) -> None:
        GraspingAction(self.object_designator, self.arm, self.grasping_prepose_distance).perform()
        ClosingMotion(self.object_designator, self.arm).perform()
        MoveGripperMotion(GripperState.OPEN, self.arm, allow_gripper_collision=True).perform()

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        """
        Check if the container is closed, this assumes that the container state can be read accurately from the
        real world.
        """
        validate_close_open(self.object_designator, self.arm, CloseAction)

    @classmethod
    @with_plan
    def description(cls, object_designator_description: Union[Iterable[ObjectDescription.Link], ObjectDescription.Link],
                    arm: Union[Iterable[Arms], Arms] = None,
                    grasping_prepose_distance: Union[
                        Iterable[float], float] = ActionConfig.grasping_prepose_distance) -> \
            PartialDesignator[Type[CloseAction]]:
        return PartialDesignator(CloseAction, object_designator=object_designator_description,
                                 arm=arm,
                                 grasping_prepose_distance=grasping_prepose_distance)


def validate_close_open(object_designator: ObjectDescription.Link, arm: Arms,
                        action_type: Union[Type[OpenAction], Type[CloseAction]]):
    """
    Validates if the container is opened or closed by checking the joint position of the container.

    :param object_designator: The object designator_description describing the object that should be opened or closed.
    :param arm: The arm that should be used for opening or closing the container.
    :param action_type: The type of the action that should be validated.
    """
    obj_part = object_designator
    obj = object_designator.parent_entity
    container_joint_name = obj.find_joint_above_link(object_designator.name)
    lower_limit, upper_limit = obj.get_joint_limits(container_joint_name)
    joint_obj: Joint = obj.joints[container_joint_name]
    if issubclass(action_type, CloseAction):
        check_closed(joint_obj, obj_part, arm, lower_limit)
    elif issubclass(action_type, OpenAction):
        check_opened(joint_obj, obj_part, arm, upper_limit)
    else:
        raise ValueError(f"Invalid action type: {action_type}")


def check_opened(joint_obj: Joint, obj_part: Link, arm: Arms, upper_limit: float):
    if joint_obj.position < upper_limit - joint_obj.acceptable_error:
        raise ContainerManipulationError(World.robot, [arm], obj_part, joint_obj,
                                         ContainerManipulationType.Opening)


def check_closed(joint_obj: Joint, obj_part: Link, arm: Arms, lower_limit: float):
    if joint_obj.position > lower_limit + joint_obj.acceptable_error:
        raise ContainerManipulationError(World.robot, [arm], obj_part, joint_obj,
                                         ContainerManipulationType.Closing)


@has_parameters
@dataclass
class GraspingAction(ActionDescription):
    """
    Grasps an object described by the given Object Designator description
    """
    object_designator: Union[Object, ObjectDescription.Link]
    """
    Object Designator for the object that should be grasped
    """
    arm: Arms
    """
    The arm that should be used to grasp
    """
    prepose_distance: float = ActionConfig.grasping_prepose_distance
    """
    The distance in meters the gripper should be at before grasping the object
    """

    def plan(self) -> None:
        object_pose = self.object_designator.pose
        lt = LocalTransformer()
        gripper_name = RobotDescription.current_robot_description.get_arm_chain(self.arm).get_tool_frame()

        object_pose_in_gripper = lt.transform_pose(object_pose,
                                                   World.robot.get_link_tf_frame(gripper_name))

        pre_grasp = object_pose_in_gripper.copy()
        pre_grasp.pose.position.x -= self.prepose_distance

        MoveTCPMotion(pre_grasp, self.arm).perform()
        MoveGripperMotion(GripperState.OPEN, self.arm).perform()

        MoveTCPMotion(object_pose, self.arm, allow_gripper_collision=True).perform()
        MoveGripperMotion(GripperState.CLOSE, self.arm, allow_gripper_collision=True).perform()

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        body = self.object_designator
        contact_links = body.get_contact_points_with_body(World.robot).get_all_bodies()
        arm_chain = RobotDescription.current_robot_description.get_arm_chain(self.arm)
        gripper_links = arm_chain.end_effector.links
        if not any([link.name in gripper_links for link in contact_links]):
            raise ObjectNotGraspedError(self.object_designator, World.robot, self.arm, None)

    @classmethod
    @with_plan
    def description(cls, object_designator: Union[Iterable[Object], Object],
                    arm: Union[Iterable[Arms], Arms] = None,
                    prepose_distance: Union[Iterable[float], float] = ActionConfig.grasping_prepose_distance) -> \
            PartialDesignator[Type[GraspingAction]]:
        return PartialDesignator(GraspingAction, object_designator=object_designator, arm=arm,
                                 prepose_distance=prepose_distance)


@has_parameters
@dataclass
class FaceAtAction(ActionDescription):
    """
    Turn the robot chassis such that is faces the ``pose`` and after that perform a look at action.
    """

    pose: PoseStamped
    """
    The pose to face 
    """
    keep_joint_states: bool = ActionConfig.face_at_keep_joint_states
    """
    Keep the joint states of the robot the same during the navigation.
    """

    def plan(self) -> None:
        # get the robot position
        robot_position = World.robot.pose

        # calculate orientation for robot to face the object
        angle = np.arctan2(robot_position.position.y - self.pose.position.y,
                           robot_position.position.x - self.pose.position.x) + np.pi
        orientation = list(quaternion_from_euler(0, 0, angle, axes="sxyz"))

        # create new robot pose
        new_robot_pose = PoseStamped.from_list(robot_position.position.to_list(), orientation)

        # turn robot
        NavigateAction(new_robot_pose, self.keep_joint_states).perform()

        # look at target
        LookAtAction(self.pose).perform()

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        # The validation will be done in the LookAtActionPerformable.perform() method so no need to validate here.
        pass

    @classmethod
    @with_plan
    def description(cls, pose: Union[Iterable[PoseStamped], PoseStamped],
                    keep_joint_states: Union[Iterable[bool], bool] = ActionConfig.face_at_keep_joint_states) -> \
            PartialDesignator[Type[FaceAtAction]]:
        return PartialDesignator(FaceAtAction, pose=pose, keep_joint_states=keep_joint_states)


@has_parameters
@dataclass
class MoveAndPickUpAction(ActionDescription):
    """
    Navigate to `standing_position`, then turn towards the object and pick it up.
    """

    standing_position: PoseStamped
    """
    The pose to stand before trying to pick up the object
    """

    object_designator: Object
    """
    The object to pick up
    """

    arm: Arms
    """
    The arm to use
    """

    grasp: Grasp
    """
    The grasp to use
    """

    keep_joint_states: bool = ActionConfig.navigate_keep_joint_states
    """
    Keep the joint states of the robot the same during the navigation.
    """

    def plan(self):
        if self.grasp == Grasp.TOP:
            grasp = GraspDescription(Grasp.FRONT, self.grasp, False)
        else:
            grasp = GraspDescription(self.grasp, None, False)
        NavigateAction(self.standing_position, self.keep_joint_states).perform()
        FaceAtAction(self.object_designator.pose, self.keep_joint_states).perform()
        PickUpAction(self.object_designator, self.arm, grasp).perform()

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        # The validation will be done in each of the atomic action perform methods so no need to validate here.
        pass

    @classmethod
    @with_plan
    def description(cls, standing_position: Union[Iterable[PoseStamped], PoseStamped],
                    object_designator: Union[Iterable[PoseStamped], PoseStamped],
                    arm: Union[Iterable[Arms], Arms] = None,
                    grasp: Union[Iterable[Grasp], Grasp] = None,
                    keep_joint_states: Union[Iterable[bool], bool] = ActionConfig.navigate_keep_joint_states) -> \
            PartialDesignator[Type[MoveAndPickUpAction]]:
        return PartialDesignator(MoveAndPickUpAction,
                                 standing_position=standing_position,
                                 object_designator=object_designator,
                                 arm=arm,
                                 grasp=grasp)


@has_parameters
@dataclass
class MoveAndPlaceAction(ActionDescription):
    """
    Navigate to `standing_position`, then turn towards the object and pick it up.
    """

    standing_position: PoseStamped
    """
    The pose to stand before trying to pick up the object
    """

    object_designator: Object
    """
    The object to pick up
    """

    target_location: PoseStamped
    """
    The location to place the object.
    """

    arm: Arms
    """
    The arm to use
    """

    keep_joint_states: bool = ActionConfig.navigate_keep_joint_states
    """
    Keep the joint states of the robot the same during the navigation.
    """

    def plan(self):
        NavigateAction(self.standing_position, self.keep_joint_states).perform()
        FaceAtAction(self.target_location, self.keep_joint_states).perform()
        PlaceAction(self.object_designator, self.target_location, self.arm).perform()

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        # The validation will be done in each of the atomic action perform methods so no need to validate here.
        pass

    @classmethod
    @with_plan
    def description(cls, standing_position: Union[Iterable[PoseStamped], PoseStamped],
                    object_designator: Union[Iterable[Object], Object],
                    target_location: Union[Iterable[PoseStamped], PoseStamped],
                    arm: Union[Iterable[Arms], Arms] = None,
                    keep_joint_states: Union[Iterable[bool], bool] = ActionConfig.navigate_keep_joint_states, ) -> \
            PartialDesignator[Type[MoveAndPlaceAction]]:
        return PartialDesignator(MoveAndPlaceAction,
                                 standing_position=standing_position,
                                 object_designator=object_designator,
                                 target_location=target_location,
                                 arm=arm)


@has_parameters
@dataclass
class SearchAction(ActionDescription):
    """
    Searches for a target object around the given location.
    """

    target_location: PoseStamped
    """
    Location around which to look for a target object.
    """

    object_type: Type[PhysicalObject]
    """
    Type of the object which is searched for.
    """

    def plan(self) -> None:
        NavigateActionDescription(
            CostmapLocation(target=self.target_location, visible_for=World.robot)).resolve().perform()

        lt = LocalTransformer()
        target_base = lt.transform_pose(self.target_location, World.robot.tf_frame)

        target_base_left = target_base.copy()
        target_base_left.pose.position.y -= 0.5

        target_base_right = target_base.copy()
        target_base_right.pose.position.y += 0.5

        plan = TryInOrderPlan(
            SequentialPlan(
                LookAtActionDescription(target_base_left),
                DetectActionDescription(DetectionTechnique.TYPES,
                                        object_designator=BelieveObject(types=[self.object_type]))),
            SequentialPlan(
                LookAtActionDescription(target_base_right),
                DetectActionDescription(DetectionTechnique.TYPES,
                                        object_designator=BelieveObject(types=[self.object_type]))),
            SequentialPlan(
                LookAtActionDescription(target_base),
                DetectActionDescription(DetectionTechnique.TYPES,
                                        object_designator=BelieveObject(types=[self.object_type]))))

        return plan.perform()

    def validate(self, result: Optional[Any] = None, max_wait_time: Optional[timedelta] = None):
        pass

    @classmethod
    @with_plan
    def description(cls, target_location: Union[Iterable[PoseStamped], PoseStamped],
                    object_type: Union[Iterable[PhysicalObject], PhysicalObject]) -> PartialDesignator[
        Type[SearchAction]]:
        return PartialDesignator(SearchAction, target_location=target_location, object_type=object_type)


MoveTorsoActionDescription = MoveTorsoAction.description
SetGripperActionDescription = SetGripperAction.description
ParkArmsActionDescription = ParkArmsAction.description
ReachToPickUpActionDescription = ReachToPickUpAction.description
PickUpActionDescription = PickUpAction.description
PlaceActionDescription = PlaceAction.description
NavigateActionDescription = NavigateAction.description
TransportActionDescription = TransportAction.description
LookAtActionDescription = LookAtAction.description
DetectActionDescription = DetectAction.description
OpenActionDescription = OpenAction.description
CloseActionDescription = CloseAction.description
GraspingActionDescription = GraspingAction.description
FaceAtActionDescription = FaceAtAction.description
MoveAndPickUpActionDescription = MoveAndPickUpAction.description
MoveAndPlaceActionDescription = MoveAndPlaceAction.description
ReleaseActionDescription = ReleaseAction.description
GripActionDescription = GripAction.description
SearchActionDescription = SearchAction.description



import dataclasses

import numpy as np
from box import Box
from typing_extensions import List, Union, Iterable, Optional, Callable, Iterator, Sequence, Dict

from .object_designator import ObjectPart
from ..costmaps import OccupancyCostmap, VisibilityCostmap, SemanticCostmap, GaussianCostmap, Costmap
from ..datastructures.enums import JointType, Arms, Grasp
from ..datastructures.partial_designator import PartialDesignator
from ..datastructures.pose import PoseStamped, GraspDescription, GraspPose
from ..datastructures.world import World, UseProspectionWorld
from ..designator import DesignatorError, LocationDesignatorDescription, ObjectDesignatorDescription
from ..failures import RobotInCollision
from ..local_transformer import LocalTransformer
from ..object_descriptors.urdf import ObjectDescription
from ..pose_generator_and_validator import PoseGenerator, visibility_validator, pose_sequence_reachability_validator, \
    collision_check
from ..robot_description import RobotDescription
from ..ros import logdebug
from ..world_concepts.world_object import Object, Link
from ..world_reasoning import link_pose_for_joint_config, contact, is_held_object, prospect_robot_contact
from ..config.action_conf import ActionConfig


class Location(LocationDesignatorDescription):
    """
    Default location designator which only wraps a pose.
    """

    def __init__(self, pose: PoseStamped):
        """
        Basic location designator that represents a single pose.

        :param pose: The pose that should be represented by this location designator
        """
        super().__init__()
        self.pose: PoseStamped = pose

    def ground(self) -> PoseStamped:
        """
        Default specialized_designators which returns a resolved designator which contains the pose given in init.

        :return: A resolved designator
        """
        return self.pose


@dataclasses.dataclass
class CostmapLocation(LocationDesignatorDescription):
    """
    Uses Costmaps to create locations for complex constrains
    """

    def __init__(self, target: Union[PoseStamped, Object],
                 reachable_for: Optional[Union[Iterable[Object], Object]] = None,
                 visible_for: Optional[Union[Iterable[Object], Object]] = None,
                 reachable_arm: Optional[Union[Iterable[Arms], Arms]] = None,
                 ignore_collision_with: Optional[Union[Iterable[Object], Object]] = None,
                 grasp_descriptions: Optional[Union[Iterable[GraspDescription], GraspDescription]] = None,
                 object_in_hand: Optional[Union[Iterable[Object], Object]] = None):
        """
        Location designator that uses costmaps as base to calculate locations for complex constrains like reachable or
        visible. In case of reachable the resolved location contains a list of arms with which the location is reachable.

        :param target: Location for which visibility or reachability should be calculated
        :param reachable_for: Object for which the reachability should be calculated, usually a robot
        :param visible_for: Object for which the visibility should be calculated, usually a robot
        :param reachable_arm: An optional arm with which the target should be reached
        :param ignore_collision_with: List of objects that should be ignored for collision checking.
        :param grasp_descriptions: List of grasps that should be tried to reach the target pose
        :param object_in_hand: Object that is currently in the hand of the robot
        """
        super().__init__()
        PartialDesignator.__init__(self, CostmapLocation, target=target, reachable_for=reachable_for,
                                   visible_for=visible_for,
                                   reachable_arm=reachable_arm if reachable_arm is not None else [
                                       Arms.LEFT, Arms.RIGHT],
                                   ignore_collision_with=ignore_collision_with if ignore_collision_with is not None else [
                                       []],
                                   grasp_descriptions=grasp_descriptions if grasp_descriptions is not None else [None],
                                   object_in_hand=object_in_hand)
        self.target: Union[PoseStamped, Object] = target
        self.reachable_for: Object = reachable_for
        self.visible_for: Object = visible_for
        self.reachable_arm: Optional[Arms] = reachable_arm
        self.ignore_collision_with = ignore_collision_with if ignore_collision_with is not None else [[]]
        self.grasps: List[Optional[Grasp]] = grasp_descriptions if grasp_descriptions is not None else [None]

    def ground(self) -> PoseStamped:
        """
        Default specialized_designators which returns the first result from the iterator of this instance.

        :return: A resolved location
        """
        return next(iter(self))

    @staticmethod
    def setup_costmaps(target: PoseStamped, robot: Object, visible_for, reachable_for) -> Costmap:
        """
        Sets up the costmaps for the given target and robot. The costmaps are merged and stored in the final_map


        """
        target_pose = PoseStamped.from_list([target.position.x, target.position.y, target.position.z],
                                  [target.orientation.x, target.orientation.y, target.orientation.z, target.orientation.w])
        ground_pose = PoseStamped.from_list(target_pose.position.to_list())
        ground_pose.position.z = 0

        occupancy = OccupancyCostmap(0.32, False, 200, 0.02, ground_pose)
        final_map = occupancy

        if visible_for:
            camera = robot.robot_description.get_default_camera()
            visible = VisibilityCostmap(camera.minimal_height, camera.maximal_height, 200, 0.02,
                                        PoseStamped.from_list(target_pose.position.to_list()), target_object=target, robot=robot)
            final_map += visible

        if reachable_for:
            gaussian = GaussianCostmap(200, 15, 0.02, ground_pose)
            final_map += gaussian

        return final_map

    @staticmethod
    def create_target_sequence(grasp_description: GraspDescription, target: Union[PoseStamped, Object], robot: Object,
                               object_in_hand: Object, reachable_arm: Arms) -> List[PoseStamped]:
        """
        Creates a sequence of poses which need to be reachable in this order

        :param grasp_description: Grasp description to be used for grasping
        :param target: The target of reachability, either a pose or an object
        :param robot: The robot that should be checked for reachability
        :param object_in_hand: An object that is held if any
        :param reachable_arm: The arm which should be checked for reachability
        :return: A list of poses that need to be reachable in this order
        """
        end_effector = robot.robot_description.get_arm_chain(reachable_arm).end_effector
        grasp_quaternion = end_effector.grasps[grasp_description]
        approach_axis = end_effector.get_approach_axis()
        if object_in_hand:
            current_target_pose = object_in_hand.attachments[
                World.robot].get_child_link_target_pose_given_parent(target)
            approach_offset_cm = object_in_hand.get_approach_offset()
        else:
            current_target_pose = target.copy() if isinstance(target, PoseStamped) else \
                target.get_grasp_pose(end_effector, grasp_description)
            current_target_pose.rotate_by_quaternion(grasp_quaternion)
            approach_offset_cm = 0.1 if isinstance(target, PoseStamped) else target.get_approach_offset()

        lift_pose = current_target_pose.copy()
        lift_pose.position.z += 0.1

        retract_pose = LocalTransformer().translate_pose_along_local_axis(current_target_pose,
                                                                          approach_axis,
                                                                          -approach_offset_cm)

        target_sequence = ([lift_pose, current_target_pose, retract_pose] if object_in_hand
                           else [retract_pose, current_target_pose, lift_pose])

        return target_sequence

    @staticmethod
    def create_allowed_collisions(ignore_collision_with: List[Object], object_in_hand: Object) -> Dict[Object, str]:
        """
        Creates a dict of object which are allowed to collide with the robot without impacting reachability

        :param ignore_collision_with: List of objects for which collision should be ignored
        :param object_in_hand: An object that the robot might hold
        :return: A dict of objects that link to their names
        """
        ignore_collision_with = [World.current_world.get_prospection_object_for_object(obj)
                                 for obj in ignore_collision_with]

        allowed_collision = {object: object.link_id_to_name[-1] for object in ignore_collision_with}
        if object_in_hand:
            prospection_object = World.current_world.get_prospection_object_for_object(object_in_hand)
            allowed_collision.update({prospection_object: prospection_object.link_id_to_name[-1]})
        return allowed_collision

    def __iter__(self) -> Iterator[PoseStamped]:
        """
        Generates positions for a given set of constrains from a costmap and returns
        them. The generation is based of a costmap which itself is the product of
        merging costmaps, each for a different purpose. In any case an occupancy costmap
        is used as the base, then according to the given constrains a visibility or
        gaussian costmap is also merged with this. Once the costmaps are merged,
        a generator generates pose candidates from the costmap. Each pose candidate
        is then validated against the constraints given by the designator if all validators
        pass the pose is considered valid and yielded.

           :yield: An instance of CostmapLocation.Location with a valid position that satisfies the given constraints
           """
        for params in self.generate_permutations():
            params_box = Box(params)
            # Target is either a pose or an object since the object is later needed for the visibility validator
            target = params_box.target.copy() if isinstance(params_box.target, PoseStamped) else params_box.target

            if params_box.visible_for or params_box.reachable_for:
                robot_object = params_box.visible_for if params_box.visible_for else params_box.reachable_for
                test_robot = World.current_world.get_prospection_object_for_object(robot_object)
            else:
                test_robot = World.current_world.robot

            allowed_collision = self.create_allowed_collisions(params_box.ignore_collision_with,
                                                               params_box.object_in_hand)

            final_map = self.setup_costmaps(target, test_robot, params_box.visible_for, params_box.reachable_for)

            with UseProspectionWorld():
                for pose_candidate in PoseGenerator(final_map, number_of_samples=600):
                    pose_candidate.position.z = 0
                    test_robot.set_pose(pose_candidate)
                    try:
                        collision_check(test_robot, allowed_collision)
                    except RobotInCollision:
                        continue

                    if not (params_box.reachable_for or params_box.visible_for):
                        yield pose_candidate
                        continue

                    if params_box.visible_for and not visibility_validator(test_robot, target):
                        continue

                    if not params_box.reachable_for:
                        yield pose_candidate
                        continue

                    grasp_descriptions = [
                        params_box.grasp_descriptions] if params_box.grasp_descriptions else target.calculate_grasp_descriptions(
                        test_robot)

                    for grasp_desc in grasp_descriptions:

                        target_sequence = self.create_target_sequence(grasp_desc, target, test_robot,
                                                                      params_box.object_in_hand,
                                                                      params_box.reachable_arm)

                        is_reachable = pose_sequence_reachability_validator(test_robot, target_sequence,
                                                                            arm=params_box.reachable_arm,
                                                                            allowed_collision=allowed_collision)
                        if is_reachable:
                            yield GraspPose(pose_candidate.pose, pose_candidate.header,
                                            arm=params_box.reachable_arm, grasp_description=grasp_desc)


class AccessingLocation(LocationDesignatorDescription):
    """
    Location designator which describes poses used for opening drawers
    """
    def __init__(self, handle: Union[ObjectDescription.Link, Iterable[ObjectDescription.Link]],
                 robot_desig: Union[Object, Iterable[Object]],
                 arm: Union[List[Arms], Arms] = None,
                 prepose_distance: float = ActionConfig.grasping_prepose_distance):
        """
        Describes a position from where a drawer can be opened. For now this position should be calculated before the
        drawer will be opened. Calculating the pose while the drawer is open could lead to problems.

        :param handle: ObjectPart designator for handle of the drawer
        :param robot_desig: Object designator for the robot which should open the drawer
        """
        super().__init__()
        PartialDesignator.__init__(self, AccessingLocation, handle=handle, robot_desig=robot_desig,
                                   arm=arm if arm is not None else [Arms.LEFT, Arms.RIGHT],
                                   prepose_distance=prepose_distance)
        self.handle: ObjectDescription.Link = handle
        self.robot: Object = robot_desig
        self.prepose_distance = prepose_distance
        self.arm = arm if arm is not None else [Arms.LEFT, Arms.RIGHT]

    def ground(self) -> PoseStamped:
        """
        Default specialized_designators for this location designator, just returns the first element from the iteration

        :return: A location designator for a pose from which the drawer can be opened
        """
        return next(iter(self))

    @staticmethod
    def adjust_map_for_drawer_opening(cost_map: Costmap, init_pose: PoseStamped, goal_pose: PoseStamped,
                                      width: float = 0.2):
        """
        Adjust the cost map for opening a drawer. This is done by removing all locations between the initial and final
        pose of the drawer/container.

        :param cost_map: Costmap that should be adjusted.
        :param init_pose: Pose of the drawer/container when it is fully closed.
        :param goal_pose: Pose of the drawer/container when it is fully opened.
        :param width: Width of the drawer/container.
        """
        motion_vector = [goal_pose.position.x - init_pose.position.x, goal_pose.position.y - init_pose.position.y,
                         goal_pose.position.z - init_pose.position.z]
        # remove locations between the initial and final pose
        motion_vector_length = np.linalg.norm(motion_vector)
        unit_motion_vector = np.array(motion_vector) / motion_vector_length
        orthogonal_vector = np.array([unit_motion_vector[1], -unit_motion_vector[0], 0])
        orthogonal_vector /= np.linalg.norm(orthogonal_vector)
        orthogonal_size = width
        map_origin_idx = cost_map.map.shape[0] // 2, cost_map.map.shape[1] // 2
        for i in range(int(motion_vector_length / cost_map.resolution)):
            for j in range(int(orthogonal_size / cost_map.resolution)):
                idx = (int(map_origin_idx[0] + i * unit_motion_vector[0] + j * orthogonal_vector[0]),
                       int(map_origin_idx[1] + i * unit_motion_vector[1] + j * orthogonal_vector[1]))
                cost_map.map[idx] = 0
                idx = (int(map_origin_idx[0] + i * unit_motion_vector[0] - j * orthogonal_vector[0]),
                       int(map_origin_idx[1] + i * unit_motion_vector[1] - j * orthogonal_vector[1]))
                cost_map.map[idx] = 0

    def setup_costmaps(self, handle: Link) -> Costmap:
        """
        Sets up the costmaps for the given handle and robot. The costmaps are merged and stored in the final_map.
        """
        ground_pose = PoseStamped.from_list(handle.pose.position.to_list())
        ground_pose.position.z = 0
        occupancy = OccupancyCostmap(distance_to_obstacle=0.25, from_ros=False, size=200, resolution=0.02,
                                     origin=ground_pose)
        final_map = occupancy

        gaussian = GaussianCostmap(200, 15, 0.02, ground_pose)
        final_map += gaussian

        return final_map

    def create_target_sequence(self, params_box: Box, final_map: Costmap) -> List[PoseStamped]:
        """
        Creates the sequence of target poses

        :param params_box:
        :param final_map:
        :return:
        """
        # Find a Joint that moves with the handle in the URDF tree
        container_joint = params_box.handle.parent_entity.find_joint_above_link(params_box.handle.name)

        init_pose = link_pose_for_joint_config(params_box.handle.parent_entity, {
            container_joint: params_box.handle.parent_entity.get_joint_limits(container_joint)[0]},
                                               params_box.handle.name)

        # Calculate the pose the handle would be in if the drawer was to be fully opened
        goal_pose = link_pose_for_joint_config(params_box.handle.parent_entity, {
            container_joint: params_box.handle.parent_entity.get_joint_limits(container_joint)[1] - 0.05},
                                               params_box.handle.name)

        # Handle position for calculating rotation of the final pose
        half_pose = link_pose_for_joint_config(params_box.handle.parent_entity, {
            container_joint: params_box.handle.parent_entity.get_joint_limits(container_joint)[1] / 1.5},
                                               params_box.handle.name)

        joint_type = params_box.handle.parent_entity.joints[container_joint].type

        if joint_type == JointType.PRISMATIC:
            self.adjust_map_for_drawer_opening(final_map, init_pose, goal_pose)

        target_sequence = [init_pose, half_pose, goal_pose]
        return target_sequence

    def __iter__(self) -> Iterator[PoseStamped]:
        """
        Creates poses from which the robot can open the drawer specified by the ObjectPart designator describing the
        handle. Poses are validated by checking if the robot can grasp the handle while the drawer is closed and if
        the handle can be grasped if the drawer is open.

        :yield: A location designator containing the pose and the arms that can be used.
        """
        for params in self.generate_permutations():
            params_box = Box(params)

            final_map = self.setup_costmaps(params_box.handle)

            target_sequence = self.create_target_sequence(params_box, final_map)
            half_pose = target_sequence[1]

            test_robot = World.current_world.get_prospection_object_for_object(params_box.robot_desig)

            with UseProspectionWorld():
                orientation_generator = lambda p, o: PoseGenerator.generate_orientation(p, half_pose)
                for pose_candidate in PoseGenerator(final_map, number_of_samples=600,
                                                    orientation_generator=orientation_generator):
                    pose_candidate.position.z = 0
                    test_robot.set_pose(pose_candidate)
                    try:
                        collision_check(test_robot, {})
                    except RobotInCollision:
                        continue

                    for arm_chain in test_robot.robot_description.get_manipulator_chains():
                        grasp = arm_chain.end_effector.grasps[GraspDescription(Grasp.FRONT, None, False)]
                        current_target_sequence = [pose.copy() for pose in target_sequence]
                        [pose.rotate_by_quaternion(grasp) for pose in current_target_sequence]

                        is_reachable = pose_sequence_reachability_validator(test_robot, current_target_sequence,
                                                                            arm=arm_chain.arm_type)
                        if is_reachable:
                            yield pose_candidate


class SemanticCostmapLocation(LocationDesignatorDescription):
    """
    Locations over semantic entities, like a table surface
    """

    def __init__(self, link_name, part_of, for_object=None, edges_only: bool = False,
                 horizontal_edges_only: bool = False, edge_size_in_meters: float = 0.06, height_offset: float = 0.0):
        """
        Creates a distribution over a link to sample poses which are on this link. Can be used, for example, to find
        poses that are on a table. Optionally an object can be given for which poses should be calculated, in that case
        the poses are calculated such that the bottom of the object is on the link.

        :param link_name: Name of the link for which a distribution should be calculated
        :param part_of: Object of which the urdf link is a part
        :param for_object: Optional object that should be placed at the found location
        :param edges_only: If True, only the edges of the link are considered
        :param horizontal_edges_only: If True, only the horizontal edges of the link are considered
        :param edge_size_in_meters: Size of the edges in meters.
        """
        super().__init__()
        PartialDesignator.__init__(self, SemanticCostmapLocation, link_name=link_name, part_of=part_of,
                                   for_object=for_object, edges_only=edges_only,
                                   horizontal_edges_only=horizontal_edges_only, edge_size_in_meters=edge_size_in_meters,
                                   height_offset=height_offset)
        self.link_name: str = link_name
        self.part_of: Object = part_of
        self.for_object: Optional[Object] = for_object
        self.edges_only: bool = edges_only
        self.horizontal_edges_only: bool = horizontal_edges_only
        self.edge_size_in_meters: float = edge_size_in_meters
        self.sem_costmap: Optional[SemanticCostmap] = None

    def ground(self) -> PoseStamped:
        """
        Default specialized_designators which returns the first element of the iterator of this instance.

        :return: A resolved location
        """
        return next(iter(self))

    def __iter__(self) -> Iterator[PoseStamped]:
        """
        Creates a costmap on top of a link of an Object and creates positions from it. If there is a specific Object for
        which the position should be found, a height offset will be calculated which ensures that the bottom of the Object
        is at the position in the Costmap and not the origin of the Object which is usually in the centre of the Object.

        :yield: An instance of SemanticCostmapLocation.Location with the found valid position of the Costmap.
        """
        for params in self.generate_permutations():
            params_box = Box(params)

            self.sem_costmap = SemanticCostmap(params_box.part_of, params_box.link_name)
            if params_box.edges_only or params_box.horizontal_edges_only:
                self.sem_costmap = self.sem_costmap.get_edges_map(params_box.edge_size_in_meters,
                                                                  horizontal_only=params_box.horizontal_edges_only)
            height_offset = params_box.height_offset
            if params_box.for_object:
                min_p, max_p = params_box.for_object.get_axis_aligned_bounding_box().get_min_max_points()
                height_offset = (max_p.z - min_p.z) / 2
            for maybe_pose in PoseGenerator(self.sem_costmap):
                maybe_pose.position.z += height_offset
                yield maybe_pose
